/*
 * Author: Kostas Kostiadis
 * Description: Contains all the member function implementations of the CAgent class
 * 				The CAgent class is the agent itself !!!
 * Created: 17/12/98
 * Last modified: 23/7/99
 */

#include "CGetParams.H" 
#include "../Agent_Others/Externs.H"
#include "CAgent.H"

// CREATEMODULES //////////////////////////////////////
void CAgent::CreateModules()
{
cout << "[--------Creating Modules---------]" << endl;
pBehaviours = new CBehaviours(this);		// The agent's behaviours
pMemory = new CMemory(this);				// The agent's memory
pSensors = new CSensors(this);				// The agent's sensors
pActuators = new CActuators(this);			// The agent's actuators
pPlayMode = new CPlayMode(this);			// The current play mode
cout << "[--------------Done!--------------]" << endl;

// If "new" fails to allocate the requested memory, it DOES NOT return a NULL
// pointer like malloc() does.  In newer compilers, it throws a bad_alloc exception.
if ( (pBehaviours == 0) || (pMemory == 0) || (pSensors == 0) || (pActuators == 0) || (pPlayMode == 0))
	Abort("ERROR: CAgent: CreateModules: Not enough memory to create agent...", 500);
}

// Default Constructor ////////////////////////////////
CAgent::CAgent()
{
}

// CONSTRUCTOR THAT TAKES CGetParams //////////////////
CAgent::CAgent(CGetParams AgentParams)
{	
pthread_mutex_init(&m_Mutex, NULL);	// Initialise mutex
CreateModules();				// Create all modules
m_RecvStep = AgentParams.GetSP_recv_step() * 1000;		// Initialise receive step (polling time) in micro-sec
pMemory->Init(AgentParams);		// Initialise memory
pBehaviours->Init(AgentParams);	// Initialise behaviours
pActuators->SetSimStep(AgentParams.GetAP_sim_step());	// Initialise simulation step in CActuators
pActuators->SetRecvStep(m_RecvStep);					// and in CActuators
m_Status = 1;					// Status is Alive
m_Time = 0;						// Time is 0

for (int i=0; i<MAXACTIONCOUNT; i++)					// Initialise all action buffers
	strncpy(m_LastAction[i], "", MAXACTIONLEN);			//
	
m_ActionCount = 0;						// And initialise action count to 0

m_NewTime = 0;					// NewTime is 0
m_StatConfDrop = AgentParams.GetAP_stat_conf_drop();	// Get stat conf drop from params
m_Epsilon = AgentParams.GetAP_epsilon();				// Get epsilon from params

strncpy(m_SB.VMMode, "normal", 7);				// Initialise field of view
strncpy(m_SB.VMQuality, "high", 5);				// 

cout << endl << "***** CREATING AGENT...Done! ******" << endl << endl;
}

// Default destructor //////////////////////////////////
CAgent::~CAgent()
{
// Free up all memory
cout << "[--------Deleting Modules---------]" << endl;
cout << "[Brain                            ]" << endl;
delete pBehaviours;										// The agent's behaviours
pBehaviours = 0;
cout << "[Memory                           ]" << endl;
pMemory->DeletePlayerArrays();							// First delete player arrays, and then
delete pMemory;											// The agent's memory
pMemory = 0;
cout << "[Eyes & Ears                      ]" << endl;
delete pSensors;										// The agent's sensors
pSensors = 0;
cout << "[Legs                             ]" << endl;
delete pActuators;										// The agent's actuators
pActuators = 0;
cout << "[Play Mode                        ]" << endl;
delete pPlayMode;										// The current play mode
pPlayMode = 0;
cout << "[--------------Done!--------------]" << endl;
cout << endl << "++++++++++ AGENT KILLED +++++++++++" << endl << endl;
}

// Lock/Unlock functions ////////////////////////////////////
// LOCK ////////////////////////////////////////////////////
void CAgent::Lock()			// Lock Agent
{
int status;					// Locking status

#ifdef DEBUG_SYNC
cout << "LOCKING Agent..." << endl;
#endif

status = pthread_mutex_lock(&m_Mutex);
if (status != 0)
	cerr << "WARNING: CAgent: Lock: Locking error..." << endl;
}

// UNLOCK //////////////////////////////////////////////////
void CAgent::Unlock()		// Unlock Agent
{
int status;					// Unlocking status

#ifdef DEBUG_SYNC
cout << "UNLOCKING Agent..." << endl;
#endif

status = pthread_mutex_unlock(&m_Mutex);
if (status != 0)
	cerr << "WARNING: CAgent: Unlock: Unlocking error..." << endl;
}

// MODULECHANGED /////////////////////////////////////////
void CAgent::ModuleChanged(CModule* theChangedModule)
{
if (theChangedModule == pBehaviours)		// If Behaviours changed
	{
	//cout << m_Time << " : Changed with action buffer:" << pBehaviours->GetActionBuffer();
	
	pActuators->Lock();											// Lock Actuators mutex
	pActuators->SetSendBuffer(pBehaviours->GetActionBuffer());	// Copy behaviour's action buffer to actuator's send buffer
	
	pBehaviours->ExecLock();									// ...We use the CActuators m_IsNewAction as predicate
	pActuators->SetIsNewAction(1);								// Set new action predicate as true
	pBehaviours->ExecUnlock();									// ... so we need to lock/unlock the Exec here
	
	pBehaviours->NewCycleLock();								// Lock NewCycle mutex
	pBehaviours->SetIsNewCycle(0);								// Set new cycle predicate as false
	pBehaviours->NewCycleUnlock();								// Unlock NewCycle mutex
	
	pActuators->Unlock();										// Unlock Actuators mutex
	
	// Behaviours mutex already locked here	
	pBehaviours->SetActionBuffer("");							// Clear old action from buffer
	// Behaviour mutex will be unlocked after this
	// After this point, WaitToExec or B2ASignal is called depending on the action
	}
else if (theChangedModule == pMemory)		// If Memory changed
	{
	}
else if (theChangedModule == pSensors)		// If Sensors changed (this happens after the new sensory info has been stored)
	{
	if (GetNewTime() > GetTime())			// If we have entered a new cycle
		{
		SetTime(GetNewTime());				// Set time to the new time
		
		// Memory mutex is already locked here
		// Behaviour mutex is already locked here
		pBehaviours->SetIsNewVisual(0);				// Set new visual data predicate as false

		Lock();										// Lock the agent to read/write m_ActionCount
		for (int i=0; i<m_ActionCount; i++)			// For all actions sent in this cycle
			pMemory->Update4Action(GetLastAction(i));	// Update the memory for our action

		pMemory->Update4NewCycle();					// Update for new cycle
		
		/*
			REMEMBER: Update4NewCycle needs to check which actions have been
			sent.  Hence we cannot clear the buffer before the call to 
			Update4NewCycle.  However internal action updates have to be done
			BEFORE Update4NewCycle.  This is because Update4NewCycle decays
			speeds and causes various commands (like turn which uses intertia)
			to use wrong values if the Update4NewCycle is called BEFORE the Update4Action.
			So although we go through the same for loop twice, and we hold the lock
			of the agent for a bit longer, WE SHOULD!
		*/
		
		for (int i=0; i<MAXACTIONCOUNT; i++)		// For all action buffers
			strncpy(m_LastAction[i], "", 2);		// Clear the corresponding buffers
			
		m_ActionCount = 0;							// Reset action counter
		Unlock();									// Unlock the agent
		
		pMemory->UpdateRelatives();					// Now that we have predicted object globals and our new global,
													// update mobile object relative positions
	
		pBehaviours->NewCycleLock();				// Lock Behaviours NewCycle mutex
		pBehaviours->SetIsNewCycle(1);				// Set new cycle predicate to true
		pBehaviours->NewCycleSignal();				// NewCycleSignal behaviours that we are on a new cycle
		pBehaviours->NewCycleUnlock();				// Unlock Behaviours NewCycle mutex
		
		pActuators->SetSaid(0);						// Reset to 0
		// Behaviour mutex will be unlocked after this
		// Memory mutex will be unlocked after this
		}
	}
else if (theChangedModule == pActuators)	// If Actuators changed
	{
	// Actuators already locked here
		
	pBehaviours->ExecLock();				// Lock Behaviours Exec mutex
	pActuators->SetIsNewAction(0);			// Set new action predicate as false
	pBehaviours->ExecSignal();				// ExecSignal behaviours that action has been executed
	pBehaviours->ExecUnlock();				// Unlock Behaviours Exec mutex
	
	// Actuators will be unlocked after this
	}
else if (theChangedModule == pPlayMode)		// If PlayMode changed
	{
	}
else
	cerr << "WARNING: ModuleChanged : What module is that?" << endl;
}

// REMOVEGHOSTS //////////////////////////////////////////////////
void CAgent::RemoveGhosts()
{
pMemory->RemoveGhosts();		// When new visual info is received from the sensors
								// right after the visual has been stored in memory,
								// this function tests for objects that should have been
								// visible but aren't.  If that is the case, it forgets
								// these objects.
}

// UPDATEGPOS ///////////////////////////////////////////////////////////////
void CAgent::UpdateGPos()
{
CGoal *pGoalL = pMemory->GetNonConstGoal('l');	// Get non const pointer to left goal here
CGoal *pGoalR = pMemory->GetNonConstGoal('r');	// Get non const pointer to right goal here

pBehaviours->Localise();				// Localise updates GPos only if localisation was succesfull
										// Also this function updates the global facing direction

// Memory mutex is already locked here (by SensorThread())
pMemory->UpdateGlobals();				// After localisation, update global co-ordinates of visible
										// mobile objects
										
// Memory mutex will be unlocked after this (by SensorThread())
}
	
// Sensors to Memory Communication Functions /////////////////////////////////////////////////////
// S2MWRITEPLAYER ////////////////////////////////////////////////////////////////////////
void CAgent::S2MWritePlayer(int Time, int UNum, char Side, float Dist, float Dir, float DistChng, float DirChng, float BFacing, float Facing) const // Write player to Memory.
{
pMemory->SetPlayer(Time, UNum, Side, Dist, Dir, DistChng, DirChng, BFacing, Facing);
}

// S2MWRITEGOAL //////////////////////////////////////////////////////////////////////////
void CAgent::S2MWriteGoal(int Time, char ID, float Dist, float Dir) const	// Write goal to Memory
{
pMemory->SetGoal(Time, ID, Dist, Dir);
}

// S2MWRITEBALL //////////////////////////////////////////////////////////////////////////
void CAgent::S2MWriteBall(int Time, float Dist, float Dir, float DistChng, float DirChng, int Close) const	// Write ball to Memory
{
pMemory->SetBall(Time, Dist, Dir, DistChng, DirChng, Close);
}

// S2MWRITEFLAG //////////////////////////////////////////////////////////////////////////
void CAgent::S2MWriteFlag(int Time, int ID, float Dist, float Dir, float DistChng, float DirChng) const	// Write flag to Memory 
{
pMemory->SetFlag(Time, ID, Dist, Dir, DistChng, DirChng);
}

// S2MWRITELINE //////////////////////////////////////////////////////////////////////////
void CAgent::S2MWriteLine(int Time, char ID, float Dist, float Dir) const	// Write line to Memory
{
pMemory->SetLine(Time, ID, Dist, Dir);
}

// S2MUPDATE4NEWVISUAL ////////////////////////////////////////////////////////////////////
void CAgent::S2MUpdate4NewCycle()		// Update Memory for each new cycle
{
pMemory->Update4NewCycle();
}

// Sensors to Behaviours Communication Functions ////////////////////////////////////////////////
// S2BWRITEISNEWVISUAL ////////////////////////////////////////////////////////////////
void CAgent::S2BWriteIsNewVisual(int IsTrue) const			// Write IsNewVisual to Behaviours
{
pBehaviours->SetIsNewVisual(IsTrue);
}

// S2BSIGNAL ///////////////////////////////////////////////////////////////////////////
void CAgent::S2BSignal() const								// Signal Behaviours condition variable
{
pBehaviours->Signal();
}

// Sensors to PlayMode Communication Functions ////////////////////////////////////////////////////
// S2PWRITEPLAYMODE ///////////////////////////////////////////////////////////////////////////////
void CAgent::S2PWritePlayMode(char* PlayMode)				// Update the current playmode
{
pPlayMode->SetPlayMode(PlayMode);
}

// PlayMode to Behaviours Communication Functions /////////////////////////////////////////////////
// P2BREADPLAYMODE ////////////////////////////////////////////////////////////////////////////////
int CAgent::P2BReadPlayMode(int Current) const				// Read play mode from PlayMode
{
return pPlayMode->GetPlayMode(Current);
}

// P2BREADSCORE //////////////////////////////////////////////////////////////////////////////////
int CAgent::P2BReadScoreDiff() const						// Read score difference from Behaviours
{
return pBehaviours->GetScoreDiff();
}

// P2BWRITESCORE /////////////////////////////////////////////////////////////////////////////////
void CAgent::P2BWriteScoreDiff(int Diff) const				// Write score difference to Behaviours
{
pBehaviours->SetScoreDiff(Diff);
}

// Memory to Behaviours Communication Functions ///////////////////////////////////////////////////
// M2BREADPLAYER //////////////////////////////////////////////////////////////////////////////////
const CPlayer* CAgent::M2BReadPlayer(int UNum, char Side) const // Return pointer to player
{
return pMemory->GetPlayer(UNum, Side);
}

// M2BREADPLAYERARRAY /////////////////////////////////////////////////////////////////////////////
const CPlayer*const* CAgent::M2BReadPlayerArray(int Far, int Friend) const	// Return pointer to array of players
{
return pMemory->GetPlayerArray(Far, Friend);
}

// M2BREADGOAL //////////////////////////////////////////////////////////////////////////////////////
const CGoal* CAgent::M2BReadGoal(char ID) const	// Return pointer to goal
{
return pMemory->GetGoal(ID);
}

// M2BREADNONCONSTGOAL ///////////////////////////////////////////////////////////////////////////
CGoal* CAgent::M2BReadNonConstGoal(char ID)	// Return pointer to non-const goal (for LookingAt??)
{
return pMemory->GetNonConstGoal(ID);
}

// M2BREADBALL ///////////////////////////////////////////////////////////////////////////////////
const CBall* CAgent::M2BReadBall() const	// Return pointer to ball
{
return pMemory->GetBall();
}

// M2BREADNONCONSTBALL //////////////////////////////////////////////////////////////////////////
CBall* CAgent::M2BReadNonConstBall()		// Return non-const pointer to ball (for CalcIntPoint)
{
return pMemory->GetNonConstBall();
}

// M2BREADFLAG ////////////////////////////////////////////////////////////////////////////////////
const CFlag* CAgent::M2BReadFlag(int ID) const			// Return pointer to flag
{
return pMemory->GetFlag(ID);
}

// M2BREADLINE ////////////////////////////////////////////////////////////////////////////////////
const CLine* CAgent::M2BReadLine(char ID) const			// Return pointer to line
{
return pMemory->GetLine(ID);
}

// Actuators to Memory Communication Functions ////////////////////////////////////////////////////
// A2MUPDATE4ACTION //////////////////////////////////////////////////////////////////////////////
void CAgent::A2MUpdate4Action(const char* sendbuffer)	// Update Memory model depending on the action just sent
{
pMemory->Update4Action(sendbuffer);
}
	
// ABORT ///////////////////////////////////////////////
void CAgent::Abort(char* Msg, int Code)	// Display error message and exit
{
cerr << endl;
cerr << "Agent: " << Msg;
cerr << endl;
DoExit(Code);
}

// DOEXIT /////////////////////////////////////////////
void CAgent::DoExit(int Code)		// Exit properly in case of an error
{
m_Status = Code;					// Update status for Agent (threads will exit)
exit(Code);							// A brute force exit will kill all threads !
}

// MAINLOOP ////////////////////////////////////////////
int CAgent::MainLoop()		// Agent's main loop
{
int PreviousPlayMode, PlayMode;

pthread_create(&m_SensorThread, NULL, StartSensors, pSensors->GetThis());
pthread_create(&m_ActuatorThread, NULL, StartActuators, pActuators->GetThis());
sleep(1);					// Make sure threades are up and running...
							// I will use a barrier here in the future
								
while (GetStatus() == 1)
	{
	pPlayMode->Lock();							// Lock PlayMode mutex
	PreviousPlayMode = pPlayMode->GetPlayMode(0);	// Get previous playmode
	PlayMode = pPlayMode->GetPlayMode(1);		// Get current playmode
	pPlayMode->Unlock();						// Unlock PlayMode mutex
	
	if (pBehaviours->GetFormationID() == -1)	// Losing (we play 334)
		{
		switch (m_UNum)
			{
			case 1:
				pBehaviours->PlayAsGoalKeeper(PreviousPlayMode, PlayMode);	// Goalie
				break;
			case 2:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Left Back
				break;
			case 3:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Center Back
				break;
			case 4:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Right Back
				break;
			case 5:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Left Mid
				break;
			case 6:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Center Mid
				break;
			case 7:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Right Mid
				break;
			case 8:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Left Wing
				break;
			case 9:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Left Forward
				break;
			case 10:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Right Forward
				break;
			case 11:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Right Wing
				break;
			default:
				cerr << GetTime() << " WARNING: CAgent: MainLoop: What uniform number is that?" << endl;
				break;
			}
		}
	else if (pBehaviours->GetFormationID() == 0)	// Normal (we play 433)
		{
		switch (m_UNum)
			{
			case 1:
				pBehaviours->PlayAsGoalKeeper(PreviousPlayMode, PlayMode);	// Goalie
				break;
			case 2:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Sweeper
				break;
			case 3:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Left Back
				break;
			case 4:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Center Back
				break;
			case 5:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Right Back
				break;
			case 6:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Left Mid
				break;
			case 7:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Center Mid
				break;
			case 8:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Right Mid
				break;
			case 9:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Left Forward
				break;
			case 10:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Center Forward
				break;
			case 11:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Right Forward
				break;
			default:
				cerr << GetTime() << " WARNING: CAgent: MainLoop: What uniform number is that?" << endl;
				break;
			}
		}
	else if (pBehaviours->GetFormationID() == 1)	// Winning (we play 532)
		{
		switch (m_UNum)
			{
			case 1:
				pBehaviours->PlayAsGoalKeeper(PreviousPlayMode, PlayMode);	// Goalie
				break;
			case 2:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Left Half-Back
				break;
			case 3:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Left Back
				break;
			case 4:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Center Back
				break;
			case 5:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Right Back
				break;
			case 6:
				pBehaviours->PlayAsDefender(PreviousPlayMode, PlayMode);	// Right Half-Back
				break;
			case 7:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Left Mid
				break;
			case 8:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Center Mid
				break;
			case 9:
				pBehaviours->PlayAsMidfielder(PreviousPlayMode, PlayMode);	// Right Mid
				break;
			case 10:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Left Forward
				break;
			case 11:
				pBehaviours->PlayAsForward(PreviousPlayMode, PlayMode);		// Right Forward
				break;
			default:
				cerr << GetTime() << " WARNING: CAgent: MainLoop: What uniform number is that?" << endl;
				break;
			}
		}
	else
		{
		cerr << GetTime() << " WARNING: CAgent: MainLoop: What formation ID is that?" << endl;
		}
	}

pthread_join(m_SensorThread, NULL);		// Wait for SensorThead to exit
pthread_join(m_ActuatorThread, NULL);	// Wait for SensorThead to exit
if ((GetUNum() > 0) && (GetUNum() < 12))							// If player initialised ok
	cerr << GetTime() << " Agent " << GetUNum() << " SHUTTING DOWN !" << endl;	// Display this error message
else																// else UNum is uninitialised so
	cerr << "Agent SHUTTING DOWN !" << endl;						// Display this messages instead
return m_Status;						// and return the exit code
}

// Display params for each object. Just for debugging. ////////////
// DISPLAYBEHAVIOURS /////////////////////////////////////
void CAgent::DisplayBehaviours() const
{
pBehaviours->Display();
}

// DISPLAYMEMORY ////////////////////////////////////////
void CAgent::DisplayMemory() const
{
pMemory->Display();
}

// DISPLAY SENSORS //////////////////////////////////////
void CAgent::DisplaySensors() const
{
pSensors->Display();
}

// DISPLAYACTUATORS /////////////////////////////////////
void CAgent::DisplayActuators() const
{
pActuators->Display();
}

// DISPLAYPLAYMODE //////////////////////////////////////
void CAgent::DisplayPlayMode() const
{
pPlayMode->Display();
}
