/*
 * Author: Kostas Kostiadis
 * Description: Contains all the member data and member function prototypes of the CAgent class
 * 				The CAgent class is the agent itself !!!
 * Created: 17/12/98
 * Last modified: 16/7/99
 */

#ifndef _CAGENT_H
#define _CAGENT_H

#include "CMediator.H"

class CGetParams;		// Need a forward declaration

#include <unistd.h>
#include "CBehaviours.H"
#include "CMemory.H"
#include "CSensors.H"
#include "CActuators.H"
#include "CPlayMode.H"

#ifdef MEMSET
#define bzero(a, b) memset(a, 0, b)
#endif

class CAgent: public CMediator
{
// Private member data ////////////////////////////////////////
private:

	pthread_mutex_t	m_Mutex;		// Agent's mutex

	int				m_Status;				// 1: Alive, rest are exit codes
	pthread_t 		m_SensorThread;			// Sensor thread ID
	pthread_t		m_ActuatorThread;		// Actuator thread ID

	CBehaviours*	pBehaviours;			// The agent's behaviours
	CMemory* 		pMemory;				// The agent's memory
	CSensors* 		pSensors;				// The agent's sensors
	CActuators* 	pActuators;				// The agent's actuators
	CPlayMode* 		pPlayMode;				// The current play mode

	int			m_Time;						// Time of last sensory input
	int			m_NewTime;					// Time of new sensory input (that hasn't been processed yet)

	char		m_LastAction[MAXACTIONCOUNT][MAXACTIONLEN];	// Actions sent to the server during the last cycle
	int			m_ActionCount;				// Number of actions sent during the last cycle
	
	char		m_TName[MAXTEAMNAMELEN];	// Team name	
	char		m_Side;						// Side
	int			m_UNum;						// Uniform number
	CVector		m_GPos;						// Global position
	float		m_Facing;					// Global facing direction
	float		m_BFacing;					// Global body facing direction (Facing - AB.HeadAngle)
	SenseBody	m_SB;						// Last sensebody info
	CVector		m_Velocity;					// My velocity

	float		m_StatConfDrop;				// Static object confidence drop per cycle (from agent.conf)
	float		m_Epsilon;					// Epsilon value used for float equality test (from agent.conf)
	int			m_RecvStep;					// Time step for acception of commands
	
// Protected member function //////////////////////////////////////
protected:
	virtual void CreateModules();	// Used to create all Modules (i.e. Sensors, Memory, etc.)

// Public member functions ////////////////////////////////////////
public:

// Constructors/Destructors /////////////////////////////////////////
	CAgent();				// Default constructor
	CAgent(CGetParams);		// Constructor that takes CGetParams
	virtual ~CAgent();		// Default destructor

// Lock/Unlock functions ////////////////////////////////////
	void Lock();			// Lock Agent
	void Unlock();			// Unlock Agent
	
// Mediator abstract methods are overriden here ///////////////////////
	virtual void ModuleChanged(CModule*);	// Signal a Module change
	virtual void RemoveGhosts();			// When new visual info is received from the sensors
											// right after the visual has been stored in memory,
											// this function tests for objects that should have been
											// visible but aren't.  If that is the case, it forgets
											// these objects.	
	virtual void UpdateGPos();				// When new visual info is received from the sensors
											// localise and update the global position in Agent, Behaviours,
											// and Memory
											// Also this function updates the global facing direction in 
											// these modules
											
	virtual void LockAgent()				// Lock the Agent mutex
	{ this->Lock(); }	
	virtual void UnlockAgent()				// Unlock the Agent mutex
	{ this->Unlock(); }
										
	virtual void LockMemory() const			// Lock the Memory mutex
	{ pMemory->Lock(); }	
	virtual void UnlockMemory() const		// Unlock the Memory mutex
	{ pMemory->Unlock(); }
	
	virtual void LockBehaviours() const		// Lock the Behaviours mutex
	{ pBehaviours->Lock(); }	
	virtual void UnlockBehaviours() const	// Unlock the Behaviours mutex
	{ pBehaviours->Unlock(); }

	virtual void LockPlayMode() const		// Lock the PlayMode mutex
	{ pPlayMode->Lock(); }	
	virtual void UnlockPlayMode() const		// Unlock the PlayMode mutex
	{ pPlayMode->Unlock(); }

	virtual void LockActuators() const		// Lock the Actuators mutex
	{ pActuators->Lock(); }
	virtual void UnlockActuators() const	// Unlock the Actuators mutex
	{ pActuators->Unlock(); }
	
// Sensors to Memory Communication Functions /////////////////////////////////////////////////////
	// Parameters are time, unum, side, dist, dir, distchng, dirchng, bfacing, facing
	virtual void S2MWritePlayer(int, int, char, float, float, float, float, float, float) const;	// Write player to Memory.
	// Parameters are time, ID, dist, dir
	virtual void S2MWriteGoal(int, char, float, float) const;				// Write goal to Memory 
	// Parameters are time, dist, dir, distchng, dirchng
	virtual void S2MWriteBall(int, float, float, float, float, int) const;		// Write ball to Memory
	// Parameters are time, ID, dist, dir, distchng, dirchng			
	virtual void S2MWriteFlag(int, int, float, float, float, float) const;	// Write flag to Memory
	// Parameters are time, ID, dist, dir
	virtual void S2MWriteLine(int, char, float, float) const;				// Write line to Memory
	// No parameters
	virtual void S2MUpdate4NewCycle();		// Update Memory for each new cycle
		
// Sensors to Behaviours Communication Functions //////////////////////////////////////////////////
	// Parameters are TRUE/FALSE
	virtual void S2BWriteIsNewVisual(int) const;			// Write IsNewVisual to Behaviours
	// No parameters
	virtual void S2BSignal() const;							// Signal Behaviours condition variable
	
// Sensors to PlayMode Communication Functions ////////////////////////////////////////////////////
	virtual void S2PWritePlayMode(char*);					// Update the current playmode

// PlayMode to Behaviours Communication Functions /////////////////////////////////////////////////
	virtual int P2BReadPlayMode(int) const;					// Read play mode from PlayMode
	virtual int P2BReadScoreDiff() const;					// Read score difference from Behaviours
	virtual void P2BWriteScoreDiff(int) const;				// Write score difference to Behaviours
			
// Memory to Behaviours Communication Functions ///////////////////////////////////////////////////
	// Parameters are UNum and Side
	virtual const CPlayer* M2BReadPlayer(int, char) const;		// Return pointer to player
	// Parameters are 0 for Foe, 1 for Friend
	virtual const CPlayer*const* M2BReadPlayerArray(int, int) const;	// Return pointer to array of players
	// Parameters are ID
	virtual const CGoal* M2BReadGoal(char) const;				// Return pointer to goal
	virtual CGoal* M2BReadNonConstGoal(char);					// Return non-const pointer to goal (for LookingAt??)
	// Parameters are none			
	virtual const CBall* M2BReadBall() const;					// Return pointer to ball
	virtual CBall* M2BReadNonConstBall();						// Return non-const pointer to ball (for CalcIntPoint)
	// Parameters are ID
	virtual const CFlag* M2BReadFlag(int) const;				// Return pointer to flag
	// Parameters are ID
	virtual const CLine* M2BReadLine(char) const;				// Return pointer to line

// Actuators to Memory Communication Functions ////////////////////////////////////////////////////
	// Parameters are sendbuffer
	virtual void A2MUpdate4Action(const char*);		// Update Memory model depending on the action just sent
				
// Error reporting functions ////////////////////////////////////////
	// Parameters are message and code
	void	Abort	(char*, int);			// Display message and exit with a given code
	// Parameters are code
	void	DoExit	(int);					// Do a proper exit in case of a critical error
		
// Basic agent attribute get/set functions //////////////////////////
// These functions are ATOMIC since data from the agent can be shared between threads //////////////////

	void	SetStatus		(int Status)		//	Set Status 
	{ m_Status = Status; }
	int		GetStatus		() const			// 	Get Status 
	{ return m_Status; }
	
	// Locks implemented before calls here /////////////////////
	void SetIsNewAction (int IsNewAction)		// Set m_IsNewAction
	{ pActuators->SetIsNewAction(IsNewAction); }
	int GetIsNewAction	() const				// Get m_IsNewAction
	{ return pActuators->GetIsNewAction(); }
	
	void	SetTime		(int Time)				//	Set Time (of last sensory input)
	{ m_Time = Time; }
	int		GetTime		() const				// 	Get Time (of last sensory input)
	{ return m_Time; }
	
	void	SetNewTime		(int NewTime)		//	Set Time (of new sensory input)
	{ m_NewTime = NewTime; }
	int		GetNewTime		() const			// 	Get Time (of new sensory input)
	{ return m_NewTime; }
	
	void	SetLastAction(int i, const char* LastAction)// Set last action
	{ strncpy(m_LastAction[i], LastAction, MAXACTIONLEN); }
	const char*	GetLastAction	(int i) const	// Get last action
	{ return m_LastAction[i]; }
	
	void	SetActionCount	(int ActionCount)	//	Set action count
	{ m_ActionCount = ActionCount; }
	int		GetActionCount		() const		// 	Get action count
	{ return m_ActionCount; }
	
	void	SetTName	(const char* TName)		// Set team name
	{ strncpy(m_TName, TName, MAXTEAMNAMELEN); }
	const char*	GetTName	() const			// Get team name
	{ return m_TName; }

	void	SetSide		(char Side)				// Set side
	{ m_Side = Side; }
	char	GetSide		() const				// Get side
	{ return m_Side; }
	
	void	SetUNum		(int UNum)				// Set uniform number
	{ m_UNum = UNum; }
	int		GetUNum		() const				// Get uniform number
	{ return m_UNum; }
	
	void	SetGPos		(CVector GPos)			// Set global position
	{ if (GPos.GetMag() >= 0) m_GPos = GPos; }	// If localisation was succesfull
	CVector	GetGPos		() const				// Get global position
	{ return m_GPos; }

	void	SetFacing	(float Facing)			// Set global facing direction
	{ m_Facing = Facing; }
	float	GetFacing	() const				// Get global facing direction
	{ return m_Facing; }
	
	void	SetBFacing	(float BFacing)			// Set global body facing direction
	{ m_BFacing = BFacing; }
	float	GetBFacing	() const				// Get global body facing direction
	{ return m_BFacing; }
	
	void	SetSB		(SenseBody SB)			// Set sense body
	{ m_SB = SB; }
	SenseBody	GetSB	() const				// Get sense body
	{ return m_SB; }

	void	SetVelocity	(CVector& V)			// Set velocity vector
	{ m_Velocity = V; }
	const CVector* GetVelocity () const			// Get velocity vector
	{ return m_Velocity.GetThis(); }
	CVector* GetNonConstVelocity()				// Get non-const velocity vector
	{ return m_Velocity.GetNonConstThis(); }

// Basic Utilities //////////////////////////////////////////////////
	float 	NormaliseAngle	(float ang)				// Normalise angle
	{ while(ang >  180.0) ang -= 360.0;
	  while(ang < -180.0) ang += 360.0;
	  return ang; }
		
// Main loop and Init/Reconnect functions //////////////////////////
	int		MainLoop	();					// Agent's main loop

	void	Init		(char*, CGetParams);	// Initialise agent and place him in the appropriate position depending on the uniform number
	void	Reconnect	(CGetParams);			// Reconnect agent and place him in the appropriate position depending on the uniform number
												// The CGetParams is passed so that the CMemory can initialise the 
												// player arrays (player size and stuff are needed)

// Accessor functions for m_Actuators and m_Sensors //////////////////////////////
	void    SetASocket	(Socket Sock)		// Store socket info to m_Actuators m_Sock
	{ pActuators->SetSocket(Sock); }
	Socket	GetASocket	() const			// Get socket info from m_Actuators m_Sock
	{ return pActuators->GetSocket(); }
	
	void	SetSSocket	(Socket Sock)		// Store socket info to m_Sensors m_Sock
	{ pSensors->SetSocket(Sock); }
	Socket  GetSSocket	() const			// Get socket info from m_Sensors m_Sock
	{ return pSensors->GetSocket(); }
	
	Socket	InitConnection	(const char* Host, int Port)// Open connection with the server
	{ return pActuators->InitConnection(Host, Port); }
	void	CloseConnection	() const					// Close connection with the server
	{ pActuators->CloseConnection(); }

	int		SendMessage		(char* Buf) const			// Send message to the server
	{ return pActuators->SendMessage(Buf); }
	int		ReceiveMessage	(char* Buf, int Size)		// Receive message from the server
	{ return pSensors->ReceiveMessage(Buf, Size); }
	
// Debugging functions /////////////////////////////////////////////
	void 	DisplayBehaviours() const;	// Display params for each object.
	void 	DisplayMemory	() const;	// Just for debugging
	void 	DisplaySensors	() const;
	void 	DisplayActuators() const;
	void 	DisplayPlayMode	() const;
	
};
 
#endif
