/*
 * Author: Kostas Kostiadis
 * Description: This is the abstract CMediator class.
 * 				This class will be inherited by the CAgent which is going to be the concrete mediator.
 * Created: 1/3/99
 * Last modified: 16/7/99
 */
 
#ifndef _CMEDIATOR_H
#define _CMEDIATOR_H

#include "../Agent_Others/GlobalParams.H"
class CModule;		// Need a forward declaration
class CPlayer;		// Need a forward declaration
class CGoal;		// Need a forward declaration
class CBall;		// Need a forward declaration
class CLine;		// Need a forward declaration
class CFlag;		// Need a forward declaration
class CVector;		// Need a forward declaration

/*
	Since this is an abstract data type (ADT) you cannot instantiate any
	objects of this class.  We can see this class is an ADT because it contains
	pure virtual functions (like CreateModule()).  Classes that inherit from this
	class need to override the pure virtual functions
	
	A Mediator is used to define an object that encapsulates how a set of objects
	interact.  Mediator promotes loose coupling by keeping objects from refering to 
	each other explicitly, and it lets you vary their interaction independently.
*/
class CMediator
{
// Protected member data //////////////////////////////////
protected:
	CMediator(){};
	virtual void CreateModules() = 0;		// Create the modules that the Mediator will communicate with
	
// Public member functions ////////////////////////////////
public:
	virtual ~CMediator(){};

	virtual void ModuleChanged(CModule*) = 0;	// See which module has changed and
												// make the appropriate changes to other modules
	virtual void RemoveGhosts() = 0;			// When new visual info is received from the sensors
												// right after the visual has been stored in memory,
												// this function tests for objects that should have been
												// visible but aren't.  If that is the case, it forgets
												// these objects.
	virtual void UpdateGPos() = 0;				// When new visual info is received from the sensors
												// localise and update the global position in Agent
												// Also this function updates the global facing direction in 
												// Agent.  Finally this function calls the CMemory::UpdateGlobals
												
	virtual void DoExit(int) = 0;				// Do a proper exit in case of a critical error

	
	virtual void LockAgent() = 0;				// Lock the Agent mutex
	virtual void UnlockAgent() = 0;				// Unlock the Agent mutex	
	virtual void LockMemory() const = 0;		// Lock the Memory mutex
	virtual void UnlockMemory() const = 0;		// Unlock the Memory mutex
	virtual void LockBehaviours() const = 0;	// Lock the Behaviours mutex
	virtual void UnlockBehaviours() const = 0;	// Unlock the Behaviours mutex
	virtual void LockPlayMode() const = 0;		// Lock the PlayMode mutex
	virtual void UnlockPlayMode() const = 0;	// Unlock the PlayMode mutex
	virtual void LockActuators() const = 0;		// Lock the Actuators mutex
	virtual void UnlockActuators() const = 0;	// Unlock the Actuators mutex
	
	virtual void SetStatus(int) = 0;			// Set agent's status: 1-Alive, rest are error codes
	virtual int GetStatus() const = 0;			// Get agent's status

	virtual void SetIsNewAction	(int) = 0;		// Set m_IsNewAction predicate
	virtual int	GetIsNewAction	() const = 0;	// Get m_IsNewAction predicate
		
	virtual void SetTime(int) = 0;				// Set time of last sensory input
	virtual int GetTime() const = 0;			// Get time of last sensory input

	virtual void SetNewTime(int) = 0;			// Set time of new sensory input (that hasn't been processed yet)
	virtual int GetNewTime() const = 0;			// Get time of new sensory input (that hasn't been processed yet)
	
	virtual void SetLastAction(int, const char*) = 0;	// Set last action sent by actuators
	virtual const char* GetLastAction(int) const = 0;	// Get last action sent by actuators

	virtual void SetActionCount(int) = 0;		// Set action count
	virtual int GetActionCount() const = 0;		// Get action count
	
	virtual void SetTName(const char*) = 0;		// Set team name
	virtual const char* GetTName() const = 0;	// Get team name
	
	virtual void SetSide(char) = 0;				// Set side
	virtual char GetSide() const = 0;			// Get side
	
	virtual void SetUNum(int) = 0;				// Set uniform number
	virtual int GetUNum() const = 0;			// Get uniform number
	
	virtual void SetGPos(CVector) = 0;			// Set global position
	virtual CVector GetGPos() const = 0;		// Get global position
	
	virtual void SetFacing(float) = 0;			// Set global facing direction
	virtual float GetFacing() const = 0;		// Get global facing direction
	
	virtual void SetBFacing(float) = 0;			// Set global body facing direction
	virtual float GetBFacing() const = 0;		// Get global body facing direction
	
	virtual void SetSB(SenseBody) = 0;			// Set sense body info
	virtual SenseBody GetSB() const = 0;		// Get sense body info															
					
	virtual void SetVelocity (CVector&) = 0;	// Set velocity vector
	virtual const CVector* GetVelocity() const = 0;	// Get velocity vector
	virtual CVector* GetNonConstVelocity() = 0;	// Get non-const velocity vector
													
// Sensors to Memory Communication Functions /////////////////////////////////////////////////////
	// Parameters are time, unum, side, dist, dir, distchng, dirchng, bfacing, facing
	virtual void S2MWritePlayer(int, int, char, float, float, float, float, float, float) const = 0; // Write player to Memory.
	// Parameters are time, ID, dist, dir
	virtual void S2MWriteGoal(int, char, float, float) const = 0;				// Write goal to Memory 
	// Parameters are time, dist, dir, distchng, dirchng
	virtual void S2MWriteBall(int, float, float, float, float, int) const = 0;	// Write ball to Memory
	// Parameters are time, ID, dist, dir, distchng, dirchng			
	virtual void S2MWriteFlag(int, int, float, float, float, float) const = 0;	// Write flag to Memory
	// Parameters are time, ID, dist, dir
	virtual void S2MWriteLine(int, char, float, float) const = 0;				// Write line to Memory
	// No parameters
	virtual void S2MUpdate4NewCycle() = 0;		// Update Memory for each new cycle	

// Sensors to Behaviours Communication Functions //////////////////////////////////////////////////
	// Parameters are TRUE/FALSE
	virtual void S2BWriteIsNewVisual(int) const = 0;			// Write IsNewVisual to Behaviours
	// No parameters
	virtual void S2BSignal() const = 0;							// Signal Behaviours condition variable
	
// Sensors to PlayMode Communication Functions ////////////////////////////////////////////////////
	virtual void S2PWritePlayMode(char*) = 0;					// Update the current playmode

// PlayMode to Behaviours Communication Functions /////////////////////////////////////////////////
	virtual int P2BReadPlayMode(int) const = 0;					// Read play mode from PlayMode
	virtual int P2BReadScoreDiff() const = 0;					// Read score difference from Behaviours
	virtual void P2BWriteScoreDiff(int) const = 0;				// Write score difference to Behaviours
	
// Memory to Behaviours Communication Functions ///////////////////////////////////////////////////
	// Parameters are UNum and Side
	virtual const CPlayer* M2BReadPlayer(int, char) const = 0;		// Return pointer to player
	// Parameters are 0 for Foe, 1 for Friend
	virtual const CPlayer*const* M2BReadPlayerArray(int, int) const = 0;	// Return pointer to array of players
	// Parameters are ID
	virtual const CGoal* M2BReadGoal(char) const = 0;				// Return pointer to goal
	virtual CGoal* M2BReadNonConstGoal(char) = 0;					// Return non-const pointer to goal (for LookingAt??)
	// Parameters are none			
	virtual const CBall* M2BReadBall() const = 0;					// Return pointer to ball
	virtual CBall* M2BReadNonConstBall() = 0;						// Return non-const pointer to ball (for CalcIntPoint)
	// Parameters are ID
	virtual const CFlag* M2BReadFlag(int) const = 0;				// Return pointer to flag
	// Parameters are ID
	virtual const CLine* M2BReadLine(char) const = 0;				// Return pointer to line

// Actuators to Memory Communication Functions ////////////////////////////////////////////////////
	// Parameters are sendbuffer
	virtual void A2MUpdate4Action(const char*) = 0;	// Update Memory model depending on the action just sent	
	
};

#endif
