/*
 * Author: Kostas Kostiadis
 * Description: This is the CModule class.  A CModule knows its Mediator.
 * 				Sensors, Behaviours, Actuators, etc. are all Modules, so they inherit this class.
 * Created: 1/3/99
 * Last modified: 16/7/99
 */
 
#ifndef _MODULE_H
#define _MODULE_H

#include "CVector.H"
#include "../Agent_Others/GlobalParams.H"
class CMediator;	// Need a forward declaration
class CPlayer;		// Need a forward declaration
class CGoal;		// Need a forward declaration
class CBall;		// Need a forward declaration
class CLine;		// Need a forward declaration
class CFlag;		// Need a forward declaration

class CModule
{
// Private member data ////////////////////////////////////////
private:
	CMediator* pMediator;

// Public member functions ////////////////////////////////////////
public:
	CModule(CMediator*);
	virtual void Changed();					// This is what a Module calls when its state changes
	virtual void RemoveGhosts();			// When new visual info is received from the sensors
											// right after the visual has been stored in memory,
											// this function tests for objects that should have been
											// visible but aren't.  If that is the case, it forgets
											// these objects.
	virtual void UpdateGPos();				// When new visual info is received from the sensors
											// localise and update the global position in Agent
											// Also this function updates the global facing direction in 
											// Agent.  Finally it calls the CMemory::UpdateGlobals
											
	virtual void DoExit(int);				// Do a proper exit in case of a critical error

	virtual void LockAgent();				// Lock the Agent mutex
	virtual void UnlockAgent();				// Unlock the Agent mutex														
	virtual void LockMemory() const;		// Lock the Memory mutex
	virtual void UnlockMemory() const;		// Unlock the Memory mutex
	virtual void LockBehaviours() const;	// Lock the Behaviours mutex
	virtual void UnlockBehaviours() const;	// Unlock the Behaviours mutex
	virtual void LockPlayMode() const;		// Lock the PlayMode mutex
	virtual void UnlockPlayMode() const;	// Unlock the PlayMode mutex
	virtual void LockActuators() const;		// Lock the Actuators mutex
	virtual void UnlockActuators() const;	// Unlock the Actuators mutex	
	
	virtual void SetStatus(int);			// Set agent's status: 1-Alive, rest are error codes
	virtual int GetStatus() const;			// Get agent's status
		
	virtual void SetIsNewAction(int);		// Set m_IsNewAction predicate
	virtual int GetIsNewAction() const;		// Get m_IsNewAction predicate
		
	virtual void SetTime(int);				// Set time of last sensory input
	virtual int GetTime() const;			// Get time of last sensory input

	virtual void SetNewTime(int);			// Set time of new sensory input (that hasn't been processed yet)
	virtual int GetNewTime() const;			// Get time of new sensory input (that hasn't been processed yet)

	virtual void SetLastAction(int, const char*);	// Set last action sent by actuators
	virtual const char* GetLastAction(int) const;	// Get last action sent by actuators
	
	virtual void SetActionCount(int);		// Set action count
	virtual int GetActionCount() const;		// Get action count
	
	virtual void SetTName(const char*);		// Set team name
	virtual const char* GetTName() const;	// Get team name
	
	virtual void SetSide(char);				// Set side
	virtual char GetSide() const;			// Get side
	
	virtual void SetUNum(int);				// Set uniform number
	virtual int GetUNum() const;			// Get uniform number
	
	virtual void SetGPos(CVector);			// Set global position
	virtual CVector GetGPos() const;		// Get global position
	
	virtual void SetFacing(float);			// Set global facing direction
	virtual float GetFacing() const;		// Get global facing direction
	
	virtual void SetBFacing(float);			// Set global body facing direction
	virtual float GetBFacing() const;		// Get global body facing direction
	
	virtual void SetSB(SenseBody);			// Set sense body info
	virtual SenseBody GetSB() const;		// Get sense body info				

	virtual void SetVelocity (CVector&);	// Set velocity vector
	virtual const CVector* GetVelocity () const;	// Get velocity vector
	virtual CVector* GetNonConstVelocity();	// Get non-const velocity vector		

// Sensors to Memory Communication Functions ///////////////////////////////////////
	// Parameters are time, unum, side, dist, dir, distchng, dirchng, bfacing, facing
	virtual void S2MWritePlayer(int, int, char, float, float, float, float, float, float) const; // Write player to Memory.
	// Parameters are time, ID, dist, dir
	virtual void S2MWriteGoal(int, char, float, float) const;				// Write goal to Memory 
	// Parameters are time, dist, dir, distchng, dirchng
	virtual void S2MWriteBall(int, float, float, float, float, int) const;	// Write ball to Memory
	// Parameters are time, ID, dist, dir, distchng, dirchng			
	virtual void S2MWriteFlag(int, int, float, float, float, float) const;	// Write flag to Memory
	// Parameters are time, ID, dist, dir
	virtual void S2MWriteLine(int, char, float, float) const;				// Write line to Memory
	// No parameters
	virtual void S2MUpdate4NewCycle();			// Update Memory for each new cycle
		
// Sensors to Behaviours Communication Functions //////////////////////////////////////////////////
	// Parameters are TRUE/FALSE
	virtual void S2BWriteIsNewVisual(int) const;			// Write IsNewVisual to Behaviours
	// No parameters
	virtual void S2BSignal() const;							// Signal Behaviours condition variable

// Sensors to PlayMode Communication Functions ////////////////////////////////////////////////////
	virtual void S2PWritePlayMode(char*);					// Update the current playmode

// PlayMode to Behaviours Communication Functions /////////////////////////////////////////////////
	virtual int P2BReadPlayMode(int) const;					// Read play mode from PlayMode
	virtual int P2BReadScoreDiff() const;					// Read score difference from Behaviours
	virtual void P2BWriteScoreDiff(int) const;				// Write score difference to Behaviours
				
// Memory to Behaviours Communication Functions ///////////////////////////////////////////////////
	// Parameters are UNum and Side
	virtual const CPlayer* M2BReadPlayer(int, char) const;		// Return pointer to player
	// Parameters are 0 for Foe, 1 for Friend
	virtual const CPlayer*const* M2BReadPlayerArray(int, int) const;	// Return pointer to array of players
	// Parameters are ID
	virtual const CGoal* M2BReadGoal(char) const;				// Return pointer to goal
	virtual CGoal* M2BReadNonConstGoal(char);					// Return non-const pointer to goal (for LookingAt??)
	// Parameters are none			
	virtual const CBall* M2BReadBall() const;					// Return pointer to ball
	virtual CBall* M2BReadNonConstBall();						// Return non-const pointer to ball (for CalcIntPoint)
	// Parameters are ID
	virtual const CFlag* M2BReadFlag(int) const;				// Return pointer to flag
	// Parameters are ID
	virtual const CLine* M2BReadLine(char) const;				// Return pointer to line

// Actuators to Memory Communication Functions ////////////////////////////////////////////////////
	// Parameters are sendbuffer
	virtual void A2MUpdate4Action(const char*);		// Update Memory model depending on the action just sent

};

#endif
