/*
 * Author: Kostas Kostiadis
 * Description: Contains all the implementations for the agent's behavioural utilities
 *				Some of these were taken from CMU98, or from Bostock & Chandler Maths book.
 * Created: 17/5/99
 * Last modified: 8/7/99
 */

#include <iostream.h>
#include "../Agent_Classes/CVector.H"
#include "../Agent_Classes/CBehaviours.H"

// SUMGEOMSERIES //////////////////////////////////////////////////////
float CBehaviours::SumGeomSeries(float X, float R, int N)			//
{
float temp = 1;						// Used to store the cummulative product

for (int i=0; i<N; i++)				// Multiply R (with itself), N times
	temp *= R;

return (X * (temp - 1) / (R - 1)); 
}

// SOLVEFORSUMGEOMSERIES //////////////////////////////////////////////
float CBehaviours::SolveForSumGeomSeries(float X, float R, float S)	// Returns -1 on error
{
float temp;

if (R < 0)
	cerr << GetTime() << "SolveForSumGeomSeries: Cannot have -ve R..." << endl;
	
temp = S * (R - 1) / X + 1;
if (temp <= 0)
	return (-1.0);
	
return (log(temp) / log(R));        
}

// LINECIRCLEINTERSECT ////////////////////////////////////////////////
// Solves the line circle intersection problem and returns the number of solutions found		
int	CBehaviours::LineCircleIntersect(CVector Origin, CVector Direction, float Radius, CVector Center, CVector* pSol1,CVector* pSol2)
{
float a,b,c,disc;
float t1, t2;
float temp;

//cout << "LCI - Origin:";
//Origin.Display();
//cout << "LCI - Direction:";
//Direction.Display();
//cout << "LCI - Radius:" << Radius << endl;
//cout << "LCI - Center:";
//Center.Display();

a = (Direction.GetX() * Direction.GetX()) + (Direction.GetY() * Direction.GetY());
b = 2.0 * ((Origin.GetX() - Center.GetX()) * Direction.GetX() + (Origin.GetY() - Center.GetY()) * Direction.GetY());
c = ((Origin.GetX() - Center.GetX()) * (Origin.GetX() - Center.GetX())) + 
((Origin.GetY() - Center.GetY()) * (Origin.GetY() - Center.GetY())) - (Radius * Radius);

disc = (b*b) - 4 * a * c;
if (disc < 0)
	{
	//cout << "LCI - Disk -ve...returning 0" << endl;
	return 0;
	}

disc = sqrt(disc);
t1 = (-b + disc) / (2.0 * a);
t2 = (-b - disc) / (2.0 * a);

if (t1 > t2)
	{
	temp = t1;
	t1 = t2;
	t2 = temp;
	}

if (t1 > 0.0)
	{
	if (t2 > 0.0)
		{
		*pSol1 = Origin + Direction * t1;
		*pSol2 = Origin + Direction * t2;
		return 2;
		}
	else
		{
		//cout << "LCI - One of the solutions -ve...returning 0" << endl;
		return 0;
		}
	}

else if (t2 > 0.0)
	{
	*pSol1 = Origin + Direction * t2;
	return 1;
	}

else
	{
	//cout << "LCI - Both solutions -ve...returning 0" << endl;
	return 0;
	}
	
//cout << "LCI - Reached end??? Returning 0" << endl;
return 0;
}

// GETCLOSESTPOINTTOLINE ///////////////////////////////////////////////////
// Calculate closest point of intersection
CVector CBehaviours::GetClosestPointToLine(CVector Point, CVector Origin, CVector Direction)
{
CVector Result(0.0, 0.0);		// Where we store the result
CVector P1, P2;					// Two points used to from the line
float A, B, C, A1, B1, C1;		// The coefficients of the two lines

if ((fabs(Direction.GetX()) < m_Epsilon) && (fabs(Direction.GetY()) < m_Epsilon))
	{
	cerr << GetTime() << " GetClosestPointToLine: Direction cannot be zero..." << endl;
	Result.SetMag(-1);			// Set error
	return Result; 				// and return
	}
	
// First construct the line from the two points /////////////////
P1 = Origin;					// Initialise the two points
P2 = Origin + Direction;		//

if (fabs(P2.GetX() - P1.GetX()) < m_Epsilon)
	{
	if (fabs(P2.GetY() - P1.GetY()) < m_Epsilon)
		{
		cerr << GetTime() << " GetClosestPointToLine: Point cannot be the same..." << endl;
		Result.SetMag(-1);			// Set error
		return Result; 				// and return
		}
	A = 1;
	B = 0;
	}
else
	{
	A = -((P2.GetY() - P1.GetY())/(P2.GetX() - P1.GetX()));
	B = 1;
	}
	
C = -(A * P2.GetX() + B * P2.GetY());

// Then project the point to the line ///////////////////////////
A1 = B;
B1 = -A;
C1 = (A * Point.GetY() - B * Point.GetX());

if (B == 0 && B1 == 0 || A/B == A1/B1)
	{
	Result.SetMag(-1);			// Set error
	return Result; 				// and return
	}
	
if (B == 0)
	{
	Result.SetX(-C/A);
	if (B1 != 0)
		{
		Result.SetY((-A1 * Result.GetX() - C1)/B1);
		Result.CalcMag();
		Result.CalcDir();
		}
	else
		Result.SetMag(-1);
		
	return Result; 
	}
	
if (B1 == 0)
	{
	Result.SetX(-C1/A1);
	if (B != 0)
		{
		Result.SetY((-A * Result.GetY() - C)/B);
		Result.CalcMag();
		Result.CalcDir();
		}
	else
		Result.SetMag(-1);
		
	return Result;
	}
	
Result.SetX((C * B1 - B * C1)/(A1 * B - A * B1));
if (B != 0)
	{
	Result.SetY((-A * Result.GetX() - C)/B);
	Result.CalcMag();
	Result.CalcDir();
	}
else
	Result.SetMag(-1);
	
return Result;
}
