/*
Michael David Seabrook

DTPass.C

Program to evaluate candidate receivers for a pass, based on a ruleset 
generated by Quinlans C4.5 decision tree learning algorithm.

Function DTBest, is used for integration into the final soccer agent.
It returns the candidate which the rules specify as 'best', modified by the 
mode string supplied.  The mode string is a string of ordered numbers, 
which define what type of receiver is requested by the robot football agent.
For example the mode could be '241', which would mean return the most 
offensive pass if it exists, else return the most wide, else return 
any receiver that has a good chance of success, else return null.



Last modified: 21/7/99
*/

#include "../Agent_Classes/CPlayer.H"
#include "../Agent_Classes/CBehaviours.H"
#include "DTPass.H"

/*
Modes are:	
	0: Most likely			1: Only positive DT
	2: Most offensive		3: Most offensive pos only
	4: Most wide			5: Most wide pos only
	6: Most centred			7: Most centred pos only
	
	The boolean is used to determine if I want the receiver
	to have a higher X co-ord relative to me
*/

const CPlayer* CBehaviours::DTBest(char* Mode, int IsForward, const CPlayer*const* carray,const CPlayer*const* farray,const CPlayer*const* fcarray,const CPlayer*const* ffarray)
{
	CVector GPos = GetGPos();			// Store global position here
	int best_player[8];
	float best_score[8], current_score, current_x, current_y;
	float example[NO_ATTRIB+1];
	int j;
	float My_X;		// passers x co-ordinate

	example[0]=-999;		// Init to error

	for(j=0;j<8;j++)
	{
		best_score[j]=-999;
		best_player[j]=-999;
	}

	My_X = -999;

	if(IsForward)
	{
		My_X=GPos.GetX();
		for(j=2;j<8;j++)
		{
			best_score[j]=My_X;
		}
	}

	if ((carray==0) || (farray==0) || (fcarray==0) || (ffarray==0))
		{
		cerr << " *** DTBest has NULL pointers as parameters... ***" << endl;
		return 0;
		}

	for(int i=0;i<22;i++)
	{
				
		if(i<11)	// exa close array
		{
			// only continue if confidence high enough.. else ignore player
			
			if(fcarray[i]==0)
				{
				cerr << " *** DTBest has NULL close friend... ***" << endl;
				return 0;
				}
				
			if(fcarray[i]->GetConf()<m_MobConfThr) continue;
				
			if((fcarray[i]->GetRMag()<MIN_PASS_DISTANCE) || (fcarray[i]->GetRMag()>MAX_PASS_DISTANCE)) continue;
			if (create_example(carray,farray,example,fcarray[i]) == 0)
				{
				cerr << " *** DTBest: create_example returned NULL... ***" << endl;
				return 0;
				}
			current_x=fcarray[i]->GetGX();
			current_y=fabs(fcarray[i]->GetGY());
		}
		else		// exa far array
		{
			if(ffarray[i-11]==0)
				{
				cerr << " *** DTBest has NULL far friend... ***" << endl;
				return 0;
				}
				
			if(ffarray[i-11]->GetConf()<m_MobConfThr) continue;
			
			if((ffarray[i-11]->GetRMag()<MIN_PASS_DISTANCE) || (ffarray[i-11]->GetRMag()>MAX_PASS_DISTANCE)) continue;
			if (create_example(carray,farray,example,ffarray[i-11]) == 0)
				{
				cerr << " *** DTBest: create_example returned NULL... ***" << endl;
				return 0;
				}
			current_x=fcarray[i-11]->GetGX();
			current_y=fabs(fcarray[i-11]->GetGY());
		}
			
		current_score=assess(example);
		//cout << "Current score = " << current_score << endl;

		if (example[0] == -999)
			continue;
			
// Ignore this candidate if, depending on the mode selected, this candidate is not as forward, wide,
// centred etc as a previous candidate

	
		// positive only
		if((current_score>0) && (current_x>My_X) && (current_score>best_score[1])) 
		{
			best_score[1]=current_score;
			best_player[1]=i;
		}

		// Most offensive
		if (current_x > best_score[2]) 
		{
			best_score[2]=current_x;
			best_player[2]=i;
		}

		// Most offensive positive only
		if ((current_x > best_score[3]) && (current_score>0))
		{
			best_score[3]=current_x;
			best_player[3]=i;
		}

		// Most wide
		if ((current_x > best_score[4]) && (current_y>CENTRE_AREA))
		{
			best_score[4]=current_x;
			best_player[4]=i;
		}

		// Most wide positive only
		if((current_x > best_score[5])  && (current_y>CENTRE_AREA)&& (current_score>0))
		{
			best_score[5]=current_x;
			best_player[5]=i;
		}

		// Most centred
		if ((current_x > best_score[6]) && (current_y<=CENTRE_AREA))
		{
			best_score[6]=current_x;
			best_player[6]=i;
		}

		// Most centred positive only
		if((current_x > best_score[7])  && (current_y<=CENTRE_AREA) && (current_score>0)) 
		{
			best_score[7]=current_x;
			best_player[7]=i;
		}
					
			
		// default

		if((current_score>best_score[0]) && (current_x>My_X)) 
		{
			best_score[0]=current_score;
			best_player[0]=i;
		}
	}

	// Return the best candidate according to the mode string supplied.  If 
	// no candidate exists for the first mode, then the second mode character is examined
	


	for(j=0;j<9;j++)
	{
		if(Mode[j]=='\0') return 0;	// no candidates found
		
		if(best_player[Mode[j]-'0']!=-999)
		{
			if(best_player[Mode[j]-'0']<11) return fcarray[best_player[Mode[j]-'0']];
			else return ffarray[best_player[Mode[j]-'0'] - 11];

		}
	}
	cerr << GetTime() << "WARNING: DTBest: Should not reach this point..." << endl;
	return 0;
}


float CBehaviours::assess(float example[])
{
	// return decision tree rules analysis of this state of play

//	Rule 1:
        if((example[dist_receiver] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent2_dist] != -999)
		&& (example[dist_receiver] <= 12.2)
        && (example[opponent1_dist] <= -1.2)
        && (example[opponent2_dist] <= -1.5))
        return 98.4;	
		//class y  [98.4%]

// Rule 2:
		else if((example[opponent1_dist] != -999)
        && (example[opponent2_dist] != -999)
        && (example[opponent3_dist] != -999)
        && (example[opponent1_dist] <= -1.2)
        && (example[opponent2_dist] <= -1.5)
        && (example[opponent3_dist] <= -3.3))
		return 97.5;
        //->  class y  [97.5%]

//Rule 37:
		else if((example[opponent1_ang] != -999)
        && (example[opponent4_ang] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent1_dist] !=-999)
        && (example[opponent1_ang] > 21)
        && (example[opponent4_ang] <= 142)
        && (example[opponent1_dist] > -1.2)
        && (example[opponent1_dist] <= 0.8))
        return 97;
		//->  class y  [97.0%]

//Rule 11:
		else if((example[opponent4_ang] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent2_dist] != -999)
        && (example[opponent2_dist] != -999)
        && (example[opponent1_conf] != -999)
        && (example[opponent_ang8] != -999)
        && (example[opponent4_ang] > 61)
        && (example[opponent1_dist] > -6.1)
        && (example[opponent1_dist] <= -1.2)
        && (example[opponent2_dist] > -1.5)
        && (example[opponent2_dist] <= 2.8)
        && (example[opponent1_conf] <= 0.8)
        && (example[opponent_ang8] <= 0))
        return 96.9;
		//->  class y  [96.9%]

//Rule 60:
		else if((example[opponent1_ang] != -999)
        && (example[opponent2_ang]  != -999)
        && (example[opponent2_ang]  != -999)
        && (example[opponent1_dist]  != -999)
        && (example[opponent3_dist]  != -999)
        && (example[opponent1_ang] > 34)
        && (example[opponent2_ang] > 47)
        && (example[opponent2_ang] <= 108)
        && (example[opponent1_dist] <= 4.9)
        && (example[opponent3_dist] > 6))
        return 96.8;
		//->  class y  [96.8%]

//Rule 26:
		else if((example[opponent2_ang] != -999)
        && (example[ opponent1_dist] != -999)
        && (example[opponent2_ang] > 33)
        && (example[ opponent1_dist] <= -1.2))
        return 96.1;
		//->  class y  [96.1%]

//Rule 66:
		else if((example[opponent1_ang] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent1_conf] != -999)
        && (example[opponent3_conf] != -999)
        && (example[opponent1_ang] > 34)
        && (example[opponent1_dist] > 4.9)
        && (example[opponent1_conf] <= 0.512)
        && (example[opponent3_conf] > 0.32768))
        return 95.6;
		//->  class y  [95.6%]

//Rule 58:
		else if((example[opponent1_ang] != -999)
        && (example[opponent3_dist] != -999)
        && (example[opponent1_ang] > 34)
        && (example[opponent3_dist] <= 6))
        return 95.1;
		//->  class y  [95.1%]

//Rule 93:
		else if((example[dir_receiver] != -999)
        && (example[dir_receiver] != -999)
        && (example[opponent_ang8] != -999)
        && (example[opponent_ang12] != -999)
        && (example[dir_receiver] > -177)
        && (example[dir_receiver] <= -159)
        && (example[opponent_ang8] > 0)
        && (example[opponent_ang12] <= 0))
        return 75.6;
		//->  class y  [75.6%]

//Rule 49:
		else if((example[dist_receiver] != -999)
        && (example[opponent1_ang] != -999)
        && (example[opponent4_ang] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent4_conf] != -999)
        && (example[opponent_ang8] != -999)
        && (example[opponent_ang12] != -999)
        && (example[opponent_ang20] != -999)
        && (example[dist_receiver] > 16.4)
        && (example[opponent1_ang] <= 34)
        && (example[opponent4_ang] > 142)
        && (example[opponent1_dist] > -1.2)
        && (example[opponent4_conf] > 0.167772)
        && (example[opponent_ang8] <= 0)
        && (example[opponent_ang12] <= 0)
        && (example[opponent_ang20] <= 0))
        return -95.3;
		//->  class n  [95.3%]

//Rule 44:
		else if((example[dist_receiver] !=-999)
        && (example[opponent1_ang] !=-999)
        && (example[opponent2_conf] !=-999)
        && (example[opponent_ang12] !=-999)
        && (example[dist_receiver] > 27.1)
        && (example[opponent1_ang] <= 34)
        && (example[opponent2_conf] > 0.512)
        && (example[opponent_ang12] <= 0))
        return -91.7;
		//->  class n  [91.7%]

//Rule 32:
		else if((example[opponent1_ang] != -999)
        && (example[opponent3_ang] != -999)
        && (example[opponent4_ang] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent_ang8] != -999)
        && (example[opponent_ang12] != -999)
        && (example[opponent1_ang] <= 8)
        && (example[opponent3_ang] > 67)
        && (example[opponent4_ang] <= 142)
        && (example[opponent1_dist] > -1.2)
        && (example[opponent_ang8] <= 0)
        && (example[opponent_ang12] <= 0))
        return -90.9;
		//->  class n  [90.9%]

//Rule 86:
		else if((example[opponent_ang4] != -999)
        && (example[opponent_ang4] > 0))
        return -86.1;
		//->  class n  [86.1%]

//Rule 15:
		else if((example[opponent2_ang] != -999)
        && (example[opponent2_dist] != -999)
        && (example[opponent1_conf] != -999)
        && (example[opponent_ang8] != -999)
        && (example[opponent_ang20] != -999)
        && (example[opponent2_ang] <= 33)
        && (example[opponent2_dist] > -1.5)
        && (example[opponent1_conf] > 0.8)
        && (example[opponent_ang8] <= 0)
        && (example[opponent_ang20] <= 0))
        return -85.7;
		//->  class n  [85.7%]

//Rule 14:
		else if((example[opponent2_ang] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent2_dist] != -999)
        && (example[opponent_ang8] != -999)
        && (example[opponent_ang12] != -999)
        && (example[opponent2_ang] <= 33)
        && (example[opponent1_dist] > -5.9)
        && (example[opponent1_dist] <= -1.2)
        && (example[opponent2_dist] > 2.8)
        && (example[opponent_ang8] <= 0)
        && (example[opponent_ang12] <= 0))
        return -81.5;
		//->  class n  [81.5%]

//Rule 38:
		else if((example[opponent1_ang] != -999)
        && (example[opponent3_ang] != -999)
        && (example[opponent1_dist] != -999)
        && (example[opponent1_ang] <= 34)
        && (example[opponent3_ang] <= 51)
        && (example[opponent1_dist] > 0.8))
        return -79.7;
		//->  class n  [79.7%]

//Rule 92:
		else if((example[opponent_ang8] != -999)
        && (example[opponent_ang8] > 0))
        return -76.6;
		//->  class n  [76.6%]

//Rule 79:
		else if((example[opponent_ang12] != -999)
        && (example[opponent_ang12] > 0))
	    return -74.7;
		//->  class n  [74.7%]

//Rule 81:
		else if((example[opponent_ang20] != -999)
        && (example[opponent_ang20] > 0))
        return -73.2;
		//->  class n  [73.2%]

//Default class: y
		else
		{
			//cout << "DEFAULT CASE!!!!!!!!!" << endl;
			return 50;
		}
}

int CBehaviours::create_example(const CPlayer*const* carray,const CPlayer*const* farray,float example[], const CPlayer* receiver)
{
	

	int i,j,k,temp;
	int nearest_angle_enemy[ENEMY];
	float rel_dist[ENEMY];
	float rel_dir[ENEMY];
	float conf_array[ENEMY];
	
	float distance, dir, conf, temp_dist, temp_dir, temp_conf;
	

	float rec_x=receiver->GetGX();
	float rec_y=receiver->GetGY();
	float rec_dist=receiver->GetRMag();
	float rec_dir=receiver->GetRDir();
	
	for (i=0; i<ENEMY; i++)
		{
		nearest_angle_enemy[i] =-1;
		rel_dist[i] = -1.0;
		rel_dir[i] = -1.0;
		conf_array[i] = -1.0;
		}
	
	if ((carray==0) || (farray==0) || (receiver==0))
		return 0;	
	
	for(i=0;i<22;i++)
	{

		if(i<11)
		{
			if (carray[i]==0)
				return 0;
				
			if(carray[i]->GetConf() < m_MobConfThr) continue;
			if(carray[i]->GetRMag() > MAX_ENEMY_DISTANCE) continue;
	
			k=i;
			temp_dist=carray[i]->GetRMag();
			temp_dir=carray[i]->GetRDir();
			distance = rec_dist-temp_dist;
			dir = fabs(rec_dir-temp_dir);
			dir = NormaliseAngle(dir);
			conf=carray[i]->GetConf();
		}
		else
		{
			if (farray[i]==0)
				return 0;
				
			if(farray[i-11]->GetConf() < m_MobConfThr) continue;
			if(farray[i]->GetRMag() > MAX_ENEMY_DISTANCE) continue;
	
			k=i;
			temp_dist=farray[i-11]->GetRMag();
			temp_dir=farray[i-11]->GetRDir();
			distance = rec_dist-temp_dist;
			dir = fabs(rec_dir-temp_dir);
			dir = NormaliseAngle(dir);
			conf=farray[i-11]->GetConf();
		}

		for(j=0;j<ENEMY;j++)
		{
			if(nearest_angle_enemy[j] == -1) 	
			{
				nearest_angle_enemy[j]=k;
				rel_dir[j]=dir;
				rel_dist[j]=distance;
				conf_array[j]=conf;
				break;
			}

			else if(dir<rel_dir[j])
			{
				temp=nearest_angle_enemy[j];
				temp_dir=rel_dir[j];
				temp_dist=rel_dist[j];
				temp_conf=conf_array[j];
				nearest_angle_enemy[j]=k;
				rel_dir[j]=dir;
				rel_dist[j]=distance;
				conf_array[j]=conf;
				k=temp;
				dir= temp_dir;
				distance=temp_dist;
			}
		}
	}


/* create attributes. attributes no
0		=	receiver distance
1		=	receiver angle
2-7		=	relative angle to enemies, sorted on closest angle
8-13	=	relative distance to enemies, sorted on angle
14-19	=	confidence of opponents position, sorted on angle
20-34	=	distribution statistics
20-25	=	for i = 1 to 6. No of enemies within 4i relative degrees of receiver
26-28	=	for i - 1 to 3. No of enemies within 8 relative degrees, but with distance of 3i
29-32	=	for i - 1 to 3. No of enemies within 16 relative degrees, but with distance of 3i
33-35	=	for i - 1 to 3. No of enemies within 24 relative degrees, but with distance of 3i
*/

	example[0] = rec_dist;
	example[1] = rec_dir;

	for(i=0;i<ENEMY;i++)
	{
		if(nearest_angle_enemy[i] ==-1)
		{
			example[2+i] = -999;
			example[8+i] = -999;
			example[14+i] = -999;
		}
		else
		{
			example[2+i]=rel_dir[i];
			example[8+i]=rel_dist[i];	
			example[14+i]=conf_array[i];
		}
	}


	for(i=20;i<NO_ATTRIB;i++)
		example[i]=0;

	for(i=0;i<ENEMY;i++)
	{
		if(nearest_angle_enemy[i] == -1) continue;

		if(rel_dist[i]>0)
		{
			if(rel_dir[i]<4) example[20]++;
			else if(rel_dir[i]<8)  
			{	
				example[21]++;
				if(rel_dist[i]>-3)	example[26]++;
				else if(rel_dist[i]>-6)	example[27]++;
				else example[28]++;
			}
			else if(rel_dir[i]<12) example[22]++;
			else if(rel_dir[i]<16) 
			{	
				example[21]++;
				if(rel_dist[i]>-3)	example[29]++;
				else if(rel_dist[i]>-6)	example[30]++;
				else example[31]++;
			}
			else if(rel_dir[i]<20) example[24]++;
			else if(rel_dir[i]<24) 
			{	
				example[21]++;
				if(rel_dist[i]>-3)	example[32]++;
				else if(rel_dist[i]>-6)	example[33]++;
				else example[34]++;
			}
		}

		

	}

	return 1;
}
