/*
 * Author: Kostas Kostiadis
 * Description: Contains all the implementations for the agent's dribbling behaviours
 *              Some of this code is based on CMU's 98 low-level behaviours
 * Created: 11/6/99
 * Last modified: 23/7/99
 */

#include <iostream.h>
#include "../Agent_Classes/CBall.H"
#include "../Agent_Classes/CPlayer.H"
#include "../Agent_Classes/CBehaviours.H"

// DRIBBLEDASH ////////////////////////////////////////////////////
int	CBehaviours::DribbleDash(CVector EndPos, float Power, float DribbleAngle)	// See if a dash would keep us in control
{
const CBall *pBall = M2BReadBall();	// Get pointer to ball data
CVector GPos = GetGPos();			// Store global position here
CVector *pVelocity = GetNonConstVelocity();	// Get pointer to my velocity

if (!BallWillBeKickable(1, Power))			// If ball not kickable in the next cycle
	{
	if (pBall->IsKickable())				// Check if ball is kickable now
		{
		//cout << GetTime() << " DD - Ball not kickable in next cycle.  DribbleKicking it now..." << endl;
		return DribbleKick(EndPos, Power, DribbleAngle);	// We'll kick it to maintain control
		}

	if ((!BallWillBeKickable(2, Power)) && (!BallWillBeKickable(3, Power)))		// If lost control
		{
		//cout << GetTime() << " DD - Lost control.  Returning 0..." << endl;
		return 0;							// return 0;
		}
	}	
	
if (WillDashBeCollision(Power))				// If dash will be a collision
	if (pBall->IsKickable())				// If ball is kickable now
		{
		//cout << GetTime() << " DD - Dash will be a collision.  DribleKicking..." << endl;
		return DribbleKick(EndPos, Power, DribbleAngle);			// Kick it
		}
	else															// else
		{
		//cout << GetTime() << " DD - Dash will be a collision but we cannot kick anyway..." << endl;
		}

//cout << GetTime() << " DD - Dashing..." << endl;
return Dash(Power);											// Dash
}

// DRIBBLEKICK ////////////////////////////////////////////////////
int	CBehaviours::DribbleKick(CVector EndPos, float Power, float DribbleAngle)	// Figure kick to keep dribbling
{
//cout << GetTime() << " IN DRIBBLEKICK" << endl;
const CBall *pBall = M2BReadBall();			// Get pointer to ball data
const float GBFacing = GetBFacing();		// Store global body facing direction here
CVector GPos = GetGPos();					// Store global position here
CVector *pVelocity = GetNonConstVelocity();	// Get pointer to my velocity
CVector temp, temp2;						// Temp vectors used to calculate stuff
CVector MyPred, BallTarget, BallTraj;		// Calculating kick vector

if ((pBall->GetConf() < m_MobConfThr) || (!pBall->IsKickable()))	// If lost ball
	{
	//cout << GetTime() << " DK - Lost control.  Returning 0..." << endl;
	return 0;								// return false
	}

// Estimate my predicted position so that I get an idle cycle, then a dash
MyPred = PredPosAfterTurn(2, 0.0, Power, 1);
temp = EndPos - GPos;						// Distance to where I want to go
temp2 = MyPred - GPos;						// Distance between predicted and current

if (temp2.GetMag() > temp.GetMag())			// If we are going to go too far,
	{
	//cout << GetTime() << " DK - We will dash too far.  Slowing down..." << endl;
	MyPred = EndPos;						// we might want to dash slower
	}

BallTarget = MyPred + Polar2CVector(m_BallDribbleDist, DribbleAngle + GBFacing);
BallTarget = Global2Relative2Me(BallTarget, GPos, GBFacing);
BallTraj = BallTarget - pBall->GetRPos();

if (WillKickBeCollision(BallTraj.GetMag() / (1 + m_BallDecay), BallTraj.GetDir()))
	{
	// Trying to interpolate changeover
	//cout << GetTime() << " DK - Kick will be a collision.  Trying to interpolate changeover..." << endl;
	BallTarget = MyPred + Polar2CVector(m_BallDribbleDist, ((pBall->GetRDir() + DribbleAngle)/2) + GBFacing);
	BallTarget = Global2Relative2Me(BallTarget, GPos, GBFacing);
	BallTraj = BallTarget - pBall->GetRPos();
	
	if (WillKickBeCollision(BallTraj.GetMag() / (1 + m_BallDecay), BallTraj.GetDir()))
		{
		//cout << GetTime() << " DK - Kick will still be a collision.  Spinning round..." << endl;
		return SpinRound(0, 0);
		}
	else
		{
		//cout << GetTime() << " DK - DribbleKicking after interpolating... with power, dir " << BallTraj.GetMag() / (1 + m_BallDecay)
		//<< ", " << BallTraj.GetDir() << endl;
	
		return StraightKick(BallTraj.GetMag() / (1 + m_BallDecay), BallTraj.GetDir(), 1.0);
		}
	}
else
	{
	//cout << GetTime() << " DK - DribbleKicking... with power, dir " << BallTraj.GetMag() / (1 + m_BallDecay)
	//<< ", " << BallTraj.GetDir() << endl;
	
	return StraightKick(BallTraj.GetMag() / (1 + m_BallDecay), BallTraj.GetDir(), 1.0);
	}
	
//cout << GetTime() << " DK - Should not reach that !!!" << endl;
return 0;					// We will never reach this point
}

// DRIBBLE ////////////////////////////////////////////////////////
// Dribble to a specified direction avoiding things if necessary
int CBehaviours::Dribble(CVector EndPos, float Power, float Buffer, float DribbleAngle, int Dodge, CVector DodgePoint)
{

/*
	Dribbling becomes slower as the DribbleAngle becomes larger.  If we are dodging
	opponents and we are forced to change the angle, make sure fabs(DribbleAngle)
	does not become larger than 80.  This results to very weak dribblekicks and
	perhaps the player should do something else at that point.
*/

const CBall *pBall = M2BReadBall();	// Get pointer to ball data
const float GBFacing = GetBFacing();// Store global body facing direction here
CVector GPos = GetGPos();			// Store global position here
CVector *pVelocity = GetNonConstVelocity();	// Get pointer to my velocity
CVector Disp;						// Used when dodging
CVector temp;						// Temp vector used to calculate stuff
float TargetAngle;
float TargetAngleError = m_MaxGoAngle;

if (pBall->GetConf() < m_MobConfThr)	// If lost ball
	{
	//cout << GetTime() << " D - Lost control !!!" << endl;
	return 0;							// return false
	}

temp = GPos - EndPos;
if (temp.GetMag() < Buffer)			// If we are close to our destination
	{
	//cout << GetTime() << " D - Within buffer of destination.  Returning 0..." << endl;
	return 0;						// return true
	}

/*
	REMEMBER: If you are dodging something you are dribbling TOWARDS the
	dodging point irrelevant of where your end dribbling point is.
	I.e. if you get the ball at (20,0) and your target point is (52,0)
	but you have enabled dodging for point (0,0), then the player
	will start dodging towards (0,0) although the target point is
	the other way.  Dodging something means dribble round it BUT
	does not ignore the dodging point if that does not affect your
	current situation (like in the example above when the dodging point
	is far away).  It is up to the higher level tactical modules to
	decide what should be dodged.
*/

if (Dodge)							// If we need to dodge something
	{
	//cout << GetTime() << " D - Dodging..." << endl;
	Disp = EndPos - GPos;							// Dodge whatever side of us the ball was 
	Disp *= m_DodgeDistBuf / Disp.GetMag();			// supposed to go...
	Disp = Disp.Rotate(SignF(DribbleAngle) * 90);
	
	temp = DodgePoint - GPos;
	
	if (temp.GetMag() < m_DribbleDodgeCloseDist)
		{
		//cout << GetTime() << " D - Sharp dodge..." << endl;
		EndPos = GPos + Disp;
		}
	else
		{
		//cout << GetTime() << " D - Normal dodge" << endl;
		EndPos = (DodgePoint + Disp - GPos)*2 + GPos;
		}
	
	TargetAngleError = m_DribbleDodgeAngleErr;
	}	
	
temp = EndPos - GPos;
TargetAngle = NormaliseAngle(temp.GetDir() - GBFacing);

if (fabs(TargetAngle) > TargetAngleError)
	{
	//cout << GetTime() << " D - We want to turn..." << endl;
	if (!BallWillBeKickable(1, 0))						// If ball not kickable in next cycle
		{
		if (pBall->IsKickable())						// and it is kickable now
			{
			//cout << GetTime() << " D - Ball will not be kickable, and it is now.  Kicking..." << endl;
			return StraightKick(0, 0, 1.0);				// kick it !
			}
		else
			{
			//cout << GetTime() << " D - Ball will not be kickable, trying to catch up.  Dashing..." << endl;
			if (!DribbleDash(EndPos, Power, DribbleAngle))
				{
				//cout << GetTime() << " D - DribbleDash returned 0.  Ball lost..." << endl;
				return 0;
				}
			}
		}
	else
		{
		//cout << GetTime() << " D - Turning to the right angle.  Turning..." << endl;
		}
				
	return Turn(TargetAngle);
	}
	
if (pBall->IsKickable())
	{
	temp = BallPredPos() - PredPosAfterDash(Power);			// Check the relative angle to the ball
	TargetAngle = NormaliseAngle(temp.GetDir() - GBFacing);	// after I dash
	//cout << GetTime() << " D - Expected Ball Angle is: " << TargetAngle << endl;
	// **** THE NEXT LINE ORIGINALLY WAS WITH 90 INSTEAD OF 80 *** //
	//if ((fabs(TargetAngle - DribbleAngle) > m_MaxDribbleExpAngle) || (fabs(TargetAngle) > 90))
	if ((fabs(TargetAngle - DribbleAngle) > m_MaxDribbleExpAngle) || (fabs(TargetAngle) > 80))	// If ball ends up behind us
		{
		//cout << GetTime() << " D - Ball will end up behind us.  DribleKicking..." << endl;
		return DribbleKick(EndPos, Power, DribbleAngle);
		}
	else
		{
		//cout << GetTime() << " D - DribbleDashing..." << endl;
		return DribbleDash(EndPos, Power, DribbleAngle);
		}
	}
else
	{
	//cout << GetTime() << " D - Ball is not kickable.  DribbleDashing..." << endl;
	return DribbleDash(EndPos, Power, DribbleAngle);
	}
	
//cout << GetTime() << " D - Should not reach that..." << endl;
return 0;		// If we reach this point...ball lost!
}

// DRIBBLE ////////////////////////////////////////////////////////
// Simpler version of the above function. No dodging is done...
int CBehaviours::Dribble(CVector EndPos, float Power, float Buffer)
{
const CBall *pBall = M2BReadBall();	// Get pointer to ball data

if (pBall->GetRDir() > 0)			// If ball on the right use m_DribbleAngle
	return Dribble(EndPos, Power, Buffer, m_DribbleNormalAngle, 0, CVector(0,0));
else								// else if ball on the left use -m_DribbleAngle
	return Dribble(EndPos, Power, Buffer, -m_DribbleNormalAngle, 0, CVector(0,0));
}

// CLEVERDRIBBLE ///////////////////////////////////////////////////
int CBehaviours::CleverDribble(CVector EndPos, float Power, float Buffer)	// High level dribbling with opponent avoidance
{
/*
	This behaviours estimates the best dribbling angle depending on the
	opponents positions.  It also determines whether we need to dodge an
	opponent and which one.
	
	Opponents with an absolute relative direction greater than 90 are 
	ignored since they are considered as NOT dangerous.
*/
CVector GPos = GetGPos();				// Store global position here
const CBall *pBall = M2BReadBall();	// Get pointer to ball data
const CPlayer* pPlayerClose = M2BReadPlayer(0, 'u');			// Get pointer to close player
const CPlayer*const* pCloseFoe = M2BReadPlayerArray(0, 0);		// Get pointer to close enemy player data
float DribbleAngle;				// We need to calculate the best dribbling angle depending on opponent positions
float DistW, DirW;				// Distance and Direction weights
float BestVote = 0;				// Each opponent votes for an angle (we need the max)
float ClosestDist = 999;		// Distance to closest opponent in front of us
int IsDodge = 0;				// Will we need to dodge anyone?
CVector DodgeVector(0,0);		// And if yes, here is the dodging vector
CVector OppRel, Target;			// Used to calculate if opponent is blocking path to target
float Res;						// Resulting angle for opponent avoidance

float MaxOppWorryDist = 1.7*m_MinOppWorryDist;		// Needs to go on params (if value <= 1, dodging is disabled)

// Initialise dribble angle depending on the direction of the ball
if (pBall->GetRDir() > 0)			// If ball on the right use m_DribbleAngle
	DribbleAngle = m_DribbleNormalAngle;
else								// else if ball on the left use -m_DribbleAngle
	DribbleAngle = -m_DribbleNormalAngle;

// Test for close player (from proximity sensors)
if (pPlayerClose->GetConf() > ((m_MobConfDrop * m_MobConfDrop) - m_Epsilon))	// if new-ish
	{
	DirW = cos((DEG2RAD*pPlayerClose->GetRDir())/2);			// Calculate direction weight
	if (pPlayerClose->GetRMag() > MaxOppWorryDist)				// Calculate distance weight
		DistW = 0;												//
	else														//
		DistW = cos(((MaxOppWorryDist - pPlayerClose->GetRMag())/MaxOppWorryDist)*(PI/2));	//
	
	if ((DirW * DistW) > BestVote)								// If higher vote found
		{
		BestVote = DirW * DistW;								// Update highest vote
		DribbleAngle = NormaliseAngle(pPlayerClose->GetRDir() + 180);	// Store new dribble angle
		}

	// Check if opponent blocking the way to our target
	OppRel = pPlayerClose->GetGPos() - GPos;
	Target = EndPos - GPos;
	Res = OppRel.GetDir() - Target.GetDir();
	// if closest opponent so far, and in front of us
	if ((ClosestDist > pPlayerClose->GetRMag()) && (fabs(Res) < 90))
		{
		ClosestDist = pPlayerClose->GetRMag();					// Update closest distance
		
		// if he is close, but not too close
		if ((ClosestDist >= m_MinOppWorryDist) && (ClosestDist < MaxOppWorryDist))
			{
			IsDodge = 1;										// Set dodge to true
			DodgeVector = pPlayerClose->GetGPos();				// Set dodging point
			}
		// else if he is way too close
		else if (ClosestDist < m_MinOppWorryDist)
			{
			return 0;											// We need to abort
			}
		}
	}
	
// Test for close enemies
for (int i=0; i<11; i++)											// for all close enemy players
	{
	if (pCloseFoe[i]->GetConf() > ((m_MobConfDrop * m_MobConfDrop) - m_Epsilon))	// if new-ish
		{
		DirW = cos((DEG2RAD*pCloseFoe[i]->GetRDir())/2);			// Calculate direction weight
		if (pCloseFoe[i]->GetRMag() > MaxOppWorryDist)				// Calculate distance weight
			DistW = 0;												//
		else														//
			DistW = cos(((MaxOppWorryDist - pCloseFoe[i]->GetRMag())/MaxOppWorryDist)*(PI/2));	//
		
		if ((DirW * DistW) > BestVote)								// If higher vote found
			{
			BestVote = DirW * DistW;								// Update highest vote
			DribbleAngle = NormaliseAngle(pCloseFoe[i]->GetRDir() + 180);	// Store new dribble angle
			}
	
		// Check if opponent blocking the way to our target
		OppRel = pCloseFoe[i]->GetGPos() - GPos;
		Target = EndPos - GPos;
		Res = OppRel.GetDir() - Target.GetDir();
		// if closest opponent so far, and in front of us
		if ((ClosestDist > pCloseFoe[i]->GetRMag()) && (fabs(Res) < 90))
			{
			ClosestDist = pCloseFoe[i]->GetRMag();					// Update closest distance
			
			// if he is close, but not too close
			if ((ClosestDist >= m_MinOppWorryDist) && (ClosestDist < MaxOppWorryDist) && 
			(GPos.GetX() < 40))
				{
				IsDodge = 1;										// Set dodge to true
				DodgeVector = pCloseFoe[i]->GetGPos();				// Set dodging point
				}
			// else if he is way too close
			else if ((ClosestDist < m_MinOppWorryDist) || 
			((GPos.GetX() > 40) && (fabs(GPos.GetY() < 10)) && (ClosestDist < 7)))
				{
				return 0;											// We need to abort
				}
			}
		}
	}

if (DribbleAngle >= 0)
	DribbleAngle = Min(DribbleAngle, m_DribbleNormalAngle);
else
	DribbleAngle = Max(DribbleAngle, -m_DribbleNormalAngle);

return Dribble(EndPos, Power, Buffer, DribbleAngle, IsDodge, DodgeVector);	// dribble
}
