/*
 * Author: Kostas Kostiadis
 * Description: Contains all the implementations for the agent's kicking behaviours
 *              Some of this code is based on CMU's 98 low-level behaviours
 * Created: 13/5/99
 * Last modified: 24/7/99
 */

#include <iostream.h>
#include "../Agent_Classes/CBall.H"
#include "../Agent_Classes/CPlayer.H"
#include "../Agent_Classes/CBehaviours.H"

// POINT2KICKVELOCITY ////////////////////////////////////////////////
/*
	In this function we specify a global point, and the velocity we
	want the ball to have when it reaches that point.  It
	does the maths, and gives us the velocity the ball should
	be kicked with, to have the desired effect
*/
float CBehaviours::Point2KickVelocity(CVector GPoint, float TargVel)
{
CVector GPos = GetGPos();				// Store global position here
CVector temp = GPoint - GPos;			// Difference between point and my position
float Steps;							// Number of steps for ball

if (TargVel < m_Epsilon)					// If target velocity near 0
	{
	return SolveForSumInfGeomSeries(m_BallDecay, temp.GetMag());
	}
else									// else
	{
	Steps = SolveForSumGeomSeries(TargVel, 1/m_BallDecay, temp.GetMag());
	return (TargVel * pow(1/m_BallDecay, Steps));
	}
	
return 1;		// We should never reach that point
}

// CANKICKSTRAIGHT ////////////////////////////////////////////////////
int CBehaviours::CanKickStraight(float EndDistance, float Direction)	// Check if we can kick the ball straight without collision
{
const CBall *pBall = M2BReadBall();	// Get pointer to ball data
// Calculate ball's trajectory
CVector BallTrajectory = Polar2CVector(EndDistance, Direction) - pBall->GetRPos();
float KickPower = CalcKickPower(EndDistance);		// Calculate power sent to the kick
CVector BallPredP = BallPredPosAfterKick(1, KickPower, Direction);	// Ball predicted position
CVector MyPredP = PredPos();										// My predicted position
CVector tmp = (BallPredP - MyPredP);								// Difference of two vectors
float angle, distance;
int result;

//cout << GetTime() << " CANKICKDIRECT: EndDistance, Direction: " << EndDistance << ", " << Direction << endl;
//cout << GetTime() << " CANKICKDIRECT: Ball RMag: " << pBall->GetRMag() << endl;
//cout << GetTime() << " CANKICKDIRECT: Ball Trajectory is: " << endl;
//BallTrajectory.Display();

/*
	Use the law of cosines to calculate the angle formed by the player, the ball's current
	position, and the ball's target position.  We use this angle to check if a straight
	kick can be achieved without collision with the player's body.
	Law of cosines:
	
		a^2 = b^2 + c^2 - 2*b*c*cosA
		
	where a,b,c are the lengths of the three sides of the triangle and A is the
	angle between the sides b and c.  Here we solve this equation wrt A.
*/
angle = RAD2DEG * acos( ((EndDistance * EndDistance) - (BallTrajectory.GetMag() * BallTrajectory.GetMag())
	- (pBall->GetRMag() * pBall->GetRMag())) / (-2 * pBall->GetRMag() * BallTrajectory.GetMag()) );

//cout << GetTime() << " CANKICKDIRECT: angle is: " << angle << endl;

if (fabs(angle) > 90)			// If |angle| > 90 then can kick straight
	{
	//cout << GetTime() << " Ok angle TRUE!" << endl;
	return 1;					// so return true
	}
	
// If we reach this point, we need to check how close to the players body the ball will go	
distance = sin(DEG2RAD*angle) * pBall->GetRMag();

//cout << GetTime() << " CANKICKDIRECT: distance is: " << distance << endl;

if (fabs(distance) > m_ClosestDist)			// Normal case
	{
	//cout << GetTime() << " Normal case TRUE!" << endl;
	return 1;
	}
else if ((tmp.GetMag() <= m_ClosestDist) && (KickPower > 60))	// Maybe ball will go through us
	{
	//cout << GetTime() << " Ball through us TRUE!" << endl;
	return 1;
	}

else										// or maybe NOT
	{
	//cout << GetTime() << " FALSE!" << endl;
	return 0;
	}

return 0;									// We will never reach this
}
	
// CALCKICKPOWER ////////////////////////////////////////////////////
float CBehaviours::CalcKickPower(float Power)
{
/*
	When kicking the ball the final power is dependent upon
	the direction and distance of the ball.  Basically the
	more powerful kick can be achieved when the ball is 
	directly in front of the player and very close to him.
	The final power is given by the following formulae:
	
	Power *= KickPowerRate*(1 - 0.25*DirDiff / 180 - 0.25*(DistBall-PlayerSize-BallSize)/KickableArea)
	
	Where KickPowerRate, PlayerSize, BallSize are specified in server.conf,
	KickableArea = PlaerSize + BallSize + KickableMargin (as specified in server.conf),
	DirDiff is the absolute value of the agle of the ball relative to the direction the player
	        is facing (i.e. if the ball is directly ahead, this should be 0), and
	DistBall is the relative distance to the ball received by the sensors.
	(in other words DirDiff = fabs(Dir) and DistBall = Dist)
	
	So the power I should kick the ball to achieve the desired result is
	
	Power / KickPowerRate*(1 - 0.25*DirDiff / 180 - 0.25*(DistBall-PlayerSize-BallSize)/KickableArea)
*/

const CBall *pBall = M2BReadBall();	// Get pointer to ball data
float ActualPower = 0.0;			// The actual power to kick the ball

ActualPower = Power / (m_KickPowerRate * (1 - 0.25*fabs(pBall->GetRDir()) / 180 - 0.25*(pBall->GetRMag()
 - m_PlayerSize - m_BallSize) / m_KickableMargin));
//cout << GetTime() << " Actual power is: " << ActualPower << endl;

if (ActualPower > m_MaxPower)		// If result greater than maxpower
	return (m_MaxPower + 0.1);		// return maxpower + 0.1.  We add 0.1 so that we can test if max power
									// has been exceeded in later stages.
else if (ActualPower < m_MinPower)	// else if less than minpower
	return (m_MinPower - 0.1);		// return minpower - 0.1. (Same as + 0.1)
else								// else
	return (ActualPower);			// return ActualPower	
}

// CALCKSPINDIRECTION /////////////////////////////////////////////////
int CBehaviours::CalcSpinDirection(float BallRDir, float TargetDir)	// Calculate spinning direction.  (Return 1:CCW, -1:CW)
{
//cout << GetTime() << " SPINDIRECTION for ballrdir, targetdir: " << BallRDir << ", " << TargetDir;
float ClockWise = TargetDir - BallRDir;
float CounterClockWise = BallRDir - TargetDir;
if (ClockWise < 0)				// If value is negative
	ClockWise += 360;			// add 360
if (CounterClockWise < 0)		// If value is negative
	CounterClockWise += 360;	// add 360
	
if (ClockWise < CounterClockWise)
	{
	//cout << " returning CW" << endl;
	return (-1);					// return ClockWise
	}
else
	{
	//cout << " returning CCW" << endl;
	return 1;					// return CounterClockWise
	}
}

// CALCKOPPAVOIDSPINDIRECTION /////////////////////////////////////////////////
// Calculate spinning direction based on nearest opponent.  (Return 1:CCW, -1:CW)
// A player that is really close (i.e. "P"layer) gets priority...
int CBehaviours::CalcOppAvoidSpinDirection(float BallRDir, float TargetDir)
{
//cout << GetTime() << " OPPAVOIDSPINDIRECTION for ballrdir, targetdir: " << BallRDir << ", " << TargetDir;
const CPlayer*const* pCloseFoe = M2BReadPlayerArray(0, 0);	// Get pointer to close enemy player data
const CPlayer* pClosePlayer = M2BReadPlayer(0, 'u');		// Get pointer to close player (i.e. "P"layer)
float mindist = 999;										// Distance to closest opponent
float dir = 0;												// Direction to closest opponent

if ((pClosePlayer != 0) && (pClosePlayer->GetConf() > m_MobConfThr))	// If player info is new
	{
	mindist = pClosePlayer->GetRMag();								// Store player's distance
	dir = pClosePlayer->GetRDir();									// and direction
	}
else																// else get closest from the array
	{
	for (int i=0; i<11; i++)												// For all close opponents
		if ((pCloseFoe[i] != 0) && (pCloseFoe[i]->GetConf() > m_MobConfThr))	// if info is new
			{
			if (mindist > pCloseFoe[i]->GetRMag())		// If this is smaller than current
				{
				mindist = pCloseFoe[i]->GetRMag();		// Store this distance
				dir = pCloseFoe[i]->GetRDir();			// and direction
				}
			}
	}
	
if (mindist < m_MinOppWorryDist)							// If distance less than threashold
	{
	if (NormaliseAngle(dir - fabs(TargetDir)) < 0)		// Depending on directions
		{
		//cout << " returning CCW" << endl;
		return 1;										// return CounterClockWise
		}	
	else
		{
		//cout << " returning CW" << endl;
		return (-1);									// return ClockWise
		}
	}
else													// else	
	{
	//cout << " no opponents to worry about...Up to CalcSpinDirection." << endl;
	return CalcSpinDirection(BallRDir, TargetDir);		// No opponents, not up to me to decide...use CalcSpinDirection
	}
	
return 1;												// We will never reach this point
}

// PLAYERMOVEMENTCORRECTION ///////////////////////////////////////////
void CBehaviours::PlayerMovementCorrection(float *Distance, float *Direction)	// Correct for movement of player (used before calling Kick)
{
CVector GPos = GetGPos();					// Store global position here
const CBall *pBall = M2BReadBall();			// Get pointer to ball data
CVector *pVelocity = GetNonConstVelocity();	// Get pointer to my velocity

//cout << GetTime()<< " PMC- CORRECTING for player movement Dist, Dir: " << *Distance << ", " << *Direction;
// Calculate the relative ball target
CVector RelativeBallTarget = pBall->GetRPos() + Polar2CVector(*Distance, *Direction);
CVector NewRelativeTarget = PredPos() + RelativeBallTarget - GPos;
CVector NewRelativeTrajectory = NewRelativeTarget - pBall->GetRPos();
*Distance = NewRelativeTrajectory.GetMag();
*Direction = NormaliseAngle(NewRelativeTrajectory.GetDir());
//cout << " after correcting Dist, Dir: " << *Distance << ", " << *Direction << endl;
}

// SPINKICK //////////////////////////////////////////////////////////
int CBehaviours::SpinKick(float EndDistance, float Direction, int StopBall, float KickFactor)	// Spin ball round body
{
//cout << GetTime() << " IN SPINKICK" << endl;
const CBall *pBall = M2BReadBall();		// Get pointer to ball data
float dist, dir, power;					// Used to store calculated results
int rotate;								// Rotation direction 1:CCW, -1:CW
CVector BallTrajectory;					// Used to store the end ball trajectory

Direction = NormaliseAngle(Direction);	// Normalise angle

if ((pBall->GetConf() < m_MobConfThr) || (!pBall->IsKickable()))	// If lost ball
	{
	//cout << GetTime() << " SR- Lost control !!!" << endl;
	return 0;							// return false
	}

if (fabs(NormaliseAngle(Direction - pBall->GetRDir())) < m_SpinDirTol)		// If within tolerance of direction
	{
	//cout << GetTime() << " SR- Within m_SpinDirTol...Ok to stop" << endl;
	// If we don't want the ball to be stopped, or it is moving slowly anyway
	if (!StopBall || (pBall->GetMag() < m_MaxStopSpeed))		
		return 1;											// return true

	dist = 0;
	dir = 0;

	PlayerMovementCorrection(&dist, &dir);					// otherwise correct for our movement
	StraightKick(dist, dir, 1.0);							// and do the kick
	return 1;												// Then return true
	}
	
// Calculate spinning direction.  If opponents nearby, avoid opponent.  Else, pick the one
// so that the ball has to travel the less distance
rotate = CalcOppAvoidSpinDirection(pBall->GetRDir(), Direction);

if (CanKickStraight(EndDistance, Direction))		// If we can kick straight without a collision
	{
	//cout << GetTime() << " SR- Can kick straight..." << endl;
	// Calculate the ball trajectory
	BallTrajectory = Polar2CVector(EndDistance, Direction) - pBall->GetRPos();
	dist = BallTrajectory.GetMag();					// and store distance
	dir = NormaliseAngle(BallTrajectory.GetDir());	// and direction
	
	power = CalcKickPower(dist);					// Convert distance to power
	power = Min(power, m_MaxSpinPower);				// make sure power is not too high
	dist = power / CalcKickPower(1);				// recalculate distance
	
	PlayerMovementCorrection(&dist, &dir);			// correct for player movement
	StraightKick(dist, dir, 1.0);					// and do the kick
	}
else if (pBall->GetRMag() < m_ClosestDist)			// if ball is too close, spin by 90 rather than 180 degrees
	{
	//cout << GetTime() << " SR- Ball too close" << endl;
	// Calculate distance and direction
	dist = 2.0 * sqrt((m_ClosestDist * m_ClosestDist) - (pBall->GetRMag() * pBall->GetRMag()));
	dir = (-90)*rotate + pBall->GetRDir();
	
	PlayerMovementCorrection(&dist, &dir);		// correct for player movement
	StraightKick(dist, dir, KickFactor);		// and do the kick
	}
else											// else spin ball round body
	{
	//cout << GetTime() << " SR- Spinning..." << endl;
	// Calculate distance and direction
	dist = sqrt((pBall->GetRMag() * pBall->GetRMag()) - (m_ClosestDist * m_ClosestDist));
	dist += sqrt((m_OptCtrlDist * m_OptCtrlDist) - (m_ClosestDist * m_ClosestDist));
	dir = 180 + pBall->GetRDir() + rotate * RAD2DEG * asin(m_ClosestDist / pBall->GetRMag());
	
	PlayerMovementCorrection(&dist, &dir);		// correct for movement of player
	StraightKick(dist, dir, KickFactor);		// and do the kick
	}

return 1;		// Return true (we did a kick)
}

// STRAIGHTKICK ////////////////////////////////////////////////////////////////////
// Kick so that ball travels distance at direction, given a distance factor
int	CBehaviours::StraightKick(float Distance, float Direction, float DistFactor)
{
//cout << GetTime() << " IN STRAIGHTKICK" << endl;
const CBall *pBall = M2BReadBall();		// Get pointer to ball data
const float GFacing = GetFacing();		// Store global facing direction here
const float GBFacing = GetBFacing();	// Store global body facing direction here
CVector GPos = GetGPos();				// Store global position here
CVector *pVelocity = GetNonConstVelocity();	// Get pointer to my velocity
float InitialSpeed;						// Ball's speed after the kick
float KickPower;						// The power that will be sent in the kick command
float KickDirection;					// The direction of the kick command
CVector tmp;							// Temporary vector used for velocity corrections in the ball's trajectory
CVector NewTraj, Sol1, Sol2, FinalKick;	// Used for the Line & Circle intersection problem
int SolNum;								// Number of solutions

InitialSpeed = Distance * DistFactor;
Direction = NormaliseAngle(Direction);			// Normalise the kick angle


if ((pBall->GetConf() < m_MobConfThr) || (!pBall->IsKickable()))	// If lost ball
	{
	return 0;
	}

if (fabs(pBall->GetMag()) > m_Epsilon)			// If ball has non zero speed
	{
	tmp = Polar2CVector(InitialSpeed, Direction) - pBall->GetRelativeVelocity(*pVelocity, GFacing);	// Correct for ball velocity
	KickDirection = NormaliseAngle(tmp.GetDir());							// Set kick direction
	KickPower = CalcKickPower(tmp.GetMag());								// and power
	}
else											// else if ball has near zero speed
	{
	KickDirection = Direction;					// Set kick direction
	KickPower = CalcKickPower(InitialSpeed);	// Calculate the kick power
	}

if (KickPower > m_MaxPower)						// If we have exceeded max power
	{
	if (fabs(pBall->GetMag()) < m_Epsilon)		// If ball has near zero speed
		{
		KickPower = m_MaxPower;					// kick as hard as possible
		}
	/*
		ELSE: Line & Circle intersection problem.  We want to kick in the right direction,
		but not as hard as desired.  If we find 2 solutions, the one we want is the second
		one, i.e. the one furthest along the line.
	*/
	else
		{
		SolNum = LineCircleIntersect(pBall->GetRPos(), 
									Polar2CVector(1, Direction),
									m_MaxPower / CalcKickPower(1), 
									Global2Relative2Me(BallPredPos(), GPos, GBFacing),
									&Sol1, &Sol2);
									
		if (SolNum == 0)		// Can't kick to desired trajectory, so just slow the ball down
			{
			KickPower = m_MaxPower;
			KickDirection = NormaliseAngle(pBall->GetDir() - GBFacing + 180);
			}
		else					// Can kick to desired trajectory, find the appropriate solution
			{
			if (SolNum == 2)	// If two solutions
				{
				NewTraj = Sol2;	// get the second one
				}
			else				// else (if 1 solution has been found)
				{
				NewTraj = Sol1;	// we get that one
				}
			
			// Work out the final kick trajectory
			FinalKick = NewTraj - Global2Relative2Me(BallPredPos(), GPos, GBFacing);
			KickPower = CalcKickPower(FinalKick.GetMag());
			KickDirection = NormaliseAngle(FinalKick.GetDir());
			}
		}	
	}

return BasicKick(KickPower,KickDirection);				// do the kick
}
