/*
 * Author: Kostas Kostiadis
 * Description: Contains all the member functions responsible for estimating
 * 				the agent's global position on the field.
 * Created: 22/1/99
 * Last modified: 13/7/99
 */

#include <iostream.h>
#include "../Agent_Classes/CLine.H"
#include "../Agent_Classes/CFlag.H"
#include "../Agent_Classes/CBehaviours.H"

// INSIDELINE ///////////////////////////////////////////////////////////////
const CLine* CBehaviours::InsideLine()	// Return the inside line if more than one is seen
{
/*
	Time-stamps are used here instead of Confidence tests since the
	most recent info needs to be obtained.  If visual arrives first
	then confidences will be 0.9.  If sense_body arrives first then
	confidences will be 1.
*/

const CLine* pLine = 0;		// Pointer to CLine
const CLine* pResult = 0;	// Store the result here
float dist = 0.0;			// If more than one lines is seen, the inside one is the one that is further away

/*
	DO NOT use (STATCONFDROP - m_Epsilon) on the confidence test because
	it will consider both 1 and STATCONFDROP confidences.
	For these calculations we need only the latest visual.
*/

pLine = M2BReadLine('l');			// For left line
if (pLine->GetConf() > (1 - m_Epsilon))		// If info on line is new
	if (pLine->GetRMag() > dist)			// If line further away (i.e. inside line)
		{
		dist = pLine->GetRMag();			// Set current distance
		pResult = pLine;					// Set Result as this line
		}
		
pLine = M2BReadLine('t');					// For top line
if (pLine->GetConf() > (1 - m_Epsilon))		// If info on line is new
	if (pLine->GetRMag() > dist)			// If line further away (i.e. inside line)
		{
		dist = pLine->GetRMag();			// Set current distance
		pResult = pLine;					// Set Result as this line
		}
		
pLine = M2BReadLine('r');					// For right line
if (pLine->GetConf() > (1 - m_Epsilon))		// If info on line is new
	if (pLine->GetRMag() > dist)			// If line further away (i.e. inside line)
		{
		dist = pLine->GetRMag();	// Set current distance
		pResult = pLine;					// Set Result as this line
		}
	
pLine = M2BReadLine('b');					// For bottom line
if (pLine->GetConf() > (1 - m_Epsilon))		// If info on line is new
	if (pLine->GetRMag() > dist)			// If line further away (i.e. inside line)
		{
		dist = pLine->GetRMag();			// Set current distance
		pResult = pLine;					// Set Result as this line
		}

return pResult;
}

// CLOSESTFLAG //////////////////////////////////////////////////////////////
const CFlag* CBehaviours::ClosestFlag()	// Returns the ID of the closest line (to use for localisation)
{
/*
	Time-stamps are used here instead of Confidence tests since the
	most recent info needs to be obtained.  If visual arrives first
	then confidences will be 0.9.  If sense_body arrives first then
	confidences will be 1.
*/

const CFlag* pFlag = 0;			// Pointer to CFlag
const CFlag* pResult = 0;		// Store the result here
float dist=999;		// If more than one flags is seen, keep here the value of the closest distance

for (int i=1; i<54; i++)
	{
	pFlag = M2BReadFlag(i);				// For each flag

	/*
		DO NOT use (STATCONFDROP - m_Epsilon) on the confidence test because
		it will consider both 1 and STATCONFDROP confidences.
		For these calculations we need only the latest visual.
	*/
	if (pFlag->GetConf() > (1 - m_Epsilon))		// If info on flag is new
		if (pFlag->GetRMag() < dist)			// If flag is closer than the one found so far
			{
			dist = pFlag->GetRMag();			// Set current distance
			pResult = pFlag;					// Set Result as this flag
			}
	}
	
return pResult;
}

// GFACEDIR ///////////////////////////////////////////////////////////////// 
float CBehaviours::GFaceDir(const CLine* pLine)		// Calculate the global facing direction given a line
{
float angle = pLine->GetGDir();				// Get line's global direction
float MyAngle;								// This is where my global facing direction is stored

/*
	Depending on which line we are using and what is the
	relative angle to that line, we can calculate our
	global facing direction.
	The global facing direction has the following format:
	
					-90
				
		180		Left Team Player		0
		
					90
					
	and
	
					90
				
		0		Right Team Player		180
		
					-90
*/
 
MyAngle = (pLine->GetRDir() < 0) ? angle - (90 + pLine->GetRDir()) : angle + (90 - pLine->GetRDir());
return (NormaliseAngle(MyAngle));
}

// LOCALISE /////////////////////////////////////////////////////////////////
void CBehaviours::Localise()	// Calculate global position given a line and a flag
{
/*
	To estimate an agent's global position on the field
	we do the following 2 things:

	First calculate the relative position from the marker which is
		RelativePosition = (flag.distance, GlobalFaceDir + flag.direction)

	Then the global position can be found using our marker like this:
		GlobalPosition = GlobalPositionOfMarker - RelativePosition

	Of course the relative position needs to be converted from polar
	to Cartesian co-ordinates to make the subtraction.
*/
float temp;					// Temp storage of global facing
CVector OldGP = GetGPos();	// ****DEBUGGING BY COMPARING POSITIONS
CVector GP = GetGPos();		// The vector to return.
sensebody SB = GetSB();		// Used to get our head angle

const CLine* pLine = InsideLine();	// The most suitable line and flag used
const CFlag* pFlag = ClosestFlag();	// for the localisation

if ( (pLine == 0) || (pFlag == 0) )		// If either pointer is NULL
	{
	#ifdef DEBUG_SENSORS
	if (GetTime())					// If time NOT equal to 0 (i.e. not before player moves into the pitch)
		cerr << "WARNING: CBehaviours: Localise: Either Line of Flag are NULL!  Unable to localise..." << endl;
	#endif
	//cout << GetTime() << "+++LINE OR FLAG ARE NULL+++ Line, Flag: " << pLine << ", " << pFlag << endl;
	GP.SetMag(-1);	// Set to error
	}
else
	{
	//cout << "\t\t\t" << GetTime() << " Line Time, Conf, ID, Dist, Dir: " << pLine->GetTime() << ", " << pLine->GetConf() 
	//<< ", " << pLine->GetID() << ", " << pLine->GetRMag() << ", " << pLine->GetRDir() << endl;
	//cout << "\t\t\t" << GetTime() << " Flag Time, Conf, ID, Dist, Dir, GX, GY: " << pFlag->GetTime() << ", " 
	//<< pFlag->GetConf() << ", " << pFlag->GetID() << ", " << pFlag->GetRMag() << ", " << pFlag->GetRDir()
	//<< ", " << pFlag->GetGX() << ", " << pFlag->GetGY() << endl;

	temp = GFaceDir(pLine);				// Calculate and store temporarily global facing direction
	SetFacing(temp);					// Set our new global facing direction
	SetBFacing(NormaliseAngle(temp - SB.HeadAngle));		// Set our new body facing direction
	// And then calculate the global position using 1 marker and the global facing direction
	GP = pFlag->GetGPos() - Polar2CVector(pFlag->GetRMag(), pFlag->GetRDir() + GetFacing());
	}

//if ((OldGP - GP).GetMag() > 3.0)
//	cout << GetTime() << " +++ POSSIBLE LOCALISATION ERROR +++ " << endl;

//cout << GetTime() << " ***LOCALISE GFacing: " << GetFacing() << " GBFacing: " << GetBFacing()
//<< " GPOS X, Y: " << GP.GetX() << ", " << GP.GetY() << " ***" << endl;
SetGPos(GP);			// Store global position (tests for -ve magnitude before storing)
}

