// File    : MCPLPosLib.cpp
//
// Purpose : Calculate safe positions to move to
// Author  : Matthew Hunter
// Version : 0.01 (16 Jun 1999)
//           0.02 (23 Jun 1999)
//                Modified MCPLPosLib init, it no longer creates
//                the MCPLPositioner object, it just determines
//                the name of the player and loads the options
//                from the configuration file.
//
//                Added two new functions MCPLFormation and
//                MCPLSpecialisation which create the
//                MCPLPositioner objects for a given formation
//                and specialisations within a formation
//                respectively.
//           0.03 (29 Jun 1999)
//                Modified MCPLPosLibInit(MCString) to allow for
//                aliased player roles
//           0.04 (6 Jul 1999)
//                New form of MCPLPosLibInit to allow the
//                configuration file to be read before connecting
//                to the soccer server.
//
//                MCPLSetShirt and MCPLSetName added to change the
//                players identity after a connection with the
//                server has been established.
//
//                MCPLPosLibLoad added to load and check the
//                configuration file.
//           0.05 (12 Jul 1999)
//                Changed to use new variants of MCPLPositioner::make
//           0.06 (14 Jul 1999)
//                Changed to use MCPLDBPosLib for debug messages
//           0.07 (20 Jul 1999)
//                Added debug message to echo contents of configuration
//                file after load.

#include "MCPLPosLib.h"
#include "MCPLPositioner.h"

#include "MCString.h"
#include "MCIni.h"

#include <stdio.h>

// Declare the initialisation file static so it does not
// pollute the global namespace.
static MCIni options;
static MCString name;
static int optionsloaded=0;

int MCPLPosLibLoad(void) {
  if (!optionsloaded) {
    optionsloaded=1;
    // Read the configuration file.
    options.import(MCString("findpos.ini"));
  }
#ifdef MCPLDBPosLib
    cout << "Dumping contents of intialisation file:" << endl;
    cout << options << endl;
#endif
  return options.get("MCPosLib","Version",-1);
}

int MCPLPosLibInit(int shirt) {
  // Need to determine the name of this player
  // from the shirt number.

  int fver=MCPLPosLibLoad();

  MCPLSetShirt(shirt);

  return fver;
}


int MCPLPosLibInit(MCString role) {
  int fver = MCPLPosLibLoad();

  MCPLSetName(role);
  return fver;
}

int MCPLPosLibInit(void) {
  return MCPLPosLibLoad();
}

void MCPLSetShirt(int shirt) {
  // Need to determine the name of this player
  // from the shirt number.

  MCString myname;

  myname=options.get("Shirts",MCString(shirt),"");
  if (myname==MCString("")) {
    char iam[16];
    sprintf(iam,"Player%02d",shirt);
    myname=MCString(iam);
  }
  MCPLSetName(myname);
}

void MCPLSetName(MCString role) {
  MCString alias,empty,Alias("Alias");
  int acount=0;
  int alimit=options.get("MCPosLib","AliasLimit",5);

  alias = options.get(role,Alias,empty);
  while((alias!=empty) && (acount<alimit)) {
#ifdef MCPLDBPosLib
    cout << "Found alias for role " << role << " = " << alias << endl;
#endif
    acount++;
    role=alias;
    alias=options.get(role,Alias,empty);
  }

  if (acount>=alimit) {
    cout << "Alias limit exceeded, check for circular references\n";
  }
  cout << "Using player name = " << role << endl;
  name = role;
}

MCPLPositioner *MCPLFormation(MCString formation) {

#ifdef MCPLDBPosLib
  cout << "Creating MCPLPositioner for role " << name;
  cout << " in formation " << formation << endl;
#endif

  return positioner->make(name,formation,options,0);
}
  
MCPLPositioner *MCPLSpecialisation(MCString formation,
				   MCString specialisation,
				   MCPLPositioner *base) {
  MCPLPositioner *retval;
  MCString spec = (formation+":")+specialisation;

#ifdef MCPLDBPosLib
  cout << "Creating MCPLPositioner for role " << name;
  cout << " with specialisation " << spec << endl;
#endif

  retval = positioner->make(name,spec,options,base);
  if (retval==0) {
    retval = base;
  }
  return retval;
}



