// File    : MCPLPosLib.h
//
// Purpose : Calculate safe positions to move to
// Author  : Matthew Hunter
// Version : 0.01 (16 Jun 1999)
//           0.02 (23 Jun 1999)
//                Modified MCPLPosLib init, it no longer creates
//                the MCPLPositioner object, it just determines
//                the name of the player and loads the options
//                from the configuration file.
//
//                Added two new functions MCPLFormation and
//                MCPLSpecialisation which create the
//                MCPLPositioner objects for a given formation
//                and specialisations within a formation
//                respectively.
//           0.03 (6 Jul 1999)
//                New form of MCPLPosLibInit to allow the
//                configuration file to be read before connecting
//                to the soccer server.
//
//                MCPLSetShirt and MCPLSetName added to change the
//                players identity after a connection with the
//                server has been established.

#ifndef MCPLPOSLIB_H
#define MCPLPOSLIB_H

#include "MCPLpoint.h"
#include "MCString.h"

//====================================================================
// The following are the external structures needed

// The MCPLStatus.h file contains the MCPLStatus structure which
// is used to pass information regarding the current state of the
// pitch to the position finder.
#include "MCPLStatus.h"

// The MCPLPositioner.h file contains the base class for all
// position finding implementations.
#include "MCPLPositioner.h"


//====================================================================
// The following are the functions that are exported

// MCPLPosLibInit
// 
// There are three forms of this function which set up
// the rest of the library ready for use.
//
// The first form takes a name role for the player.
// This will be used to extract the necessary information
// from the configuration file. This enables the function
// to determine which positioning method should be used,
// and what parameters should be used.
//
// The second form takes an integer (normally between 1 and 11
// indicating the player or 'shirt' number. This is then used
// to build a role identifier.
// e.g.
// MCPLPosLibInit(1);
// Will prepare for player with shirt number 1.
// The configuration for this player will be stored with the
// in the 'Player01' section of the configuration file.
//
// The third form takes no arguments, and is used to initialise
// and load the configuration file before the player name or shirt
// number is used. When using this form, MCPLSetShirt or MCPLSetName
// must be called before calling MCPLFormation
//
// All three variants return the version number from the configuration
// file, or -1 is no version number was found (either an old file, or
// unable to load the file etc.)
int MCPLPosLibInit(MCString);
int MCPLPosLibInit(int);
int MCPLPosLibInit();


// MCPLSetShirt
//
// This function sets the player name based on the given shirt number.
// The number to name conversion is provided by the Shirts section
// of the ini file, or a direct conversion to PlayerXX where XX is the
// 0 padded two digit shirt number.
//
// The name will then be checked for valid aliases to find the actual
// name to use.
void MCPLSetShirt(int);

// MCPLSetName
//
// This function sets the player name to the name given, which is then
// checked for aliases
void MCPLSetName(MCString);


// MCPLFormation
//
// This takes the name of a formation and prepares a 
// positioner object for the player in the formation.
MCPLPositioner *MCPLFormation(MCString);

// MCPLSpecialisation
//
// This takes the name of a formation, a specialisation of
// that formation (e.g. the name of a set play)
// and the base positioner to use if there is no
// specialisation for this player.
MCPLPositioner *MCPLSpecialisation(MCString, MCString, MCPLPositioner *);

#endif
